import os

from bamboo.wrappers.ligand_builders import BuilderObject

class GradeObject(BuilderObject):

    def _custom_init(self):
        self.program = 'grade'
        self.name = 'grade'

    def _create_program_arguments(self, ligsmile, outdir, outfile, flags):
        """Prepare the arguments for Grade to build the ligand"""

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = [ligsmile,'-name',self.outtemplate,'-opdb',self.outpdb,'-ocif',self.outcif,'-f','-resname',self.ligname]
        cmd_line_args.extend(cmd_line_flags)
        # Form Standard Input Args
        std_inpt_args = []
        cmd_line_args.extend(std_inpt_flags)

        return cmd_line_args, std_inpt_args

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = []
        std_inpt_flags = []

        if 'noHydrogens' in flags:
            cmd_line_flags.append('-really_noH')
        if 'mogul' not in flags:
            cmd_line_flags.append('-nomogul')

        return cmd_line_flags, std_inpt_flags




def residues_via_conformers(pdb_hierarchy):
    """Iterate hierarchy and yield pure residues from conformers of chains (chain->conformer->residue)"""
    for chain in pdb_hierarchy.deep_copy().chains():
        for conf in chain.conformers():
            for res in conf.residues():
                yield res

def conformers_via_residue_groups(pdb_hierarchy):
    """Iterate hierarchy and yield pure conformers from residue groups (chain->residue_group->conf)"""
    for chain in pdb_hierarchy.deep_copy().chains():
        for rg in chain.residue_groups():
            for conf in rg.conformers():
                yield conf

########################################################################################

def generate_residue_triplets(pdb_hierarchy):
    """Generate groups of three residues (this is an iterator)"""
    for chain in pdb_hierarchy.deep_copy().chains():
        # Skip if not protein
        if not chain.is_protein(): continue
        for conf in chain.conformers():
            # Reset for each conformer
            prev = current = next = None
            for res in conf.residues():
                # Shuffle all variables down one
                prev = current
                current = next
                next = res
                # Do we have three residues?
                if not (prev and current and next):
                    continue
                # Are they continuous residues
                if not (prev.resseq_as_int()+1 == current.resseq_as_int() == next.resseq_as_int()-1):
                    continue
                # Return the residue triplet
                yield (prev, current, next)



import os, sys, glob, re
import time

import numpy

from cctbx import maptbx
from cctbx import crystal

from scitbx.array_family import flex

from giant.grid.utils import get_grid_points_within_distance_cutoff_of_cart_sites_2
from giant.grid.utils import get_grid_points_within_distance_cutoff_of_origin, combine_grid_point_and_grid_vectors

class atomic_mask(object):
    def __init__(self, cart_sites, grid_size, unit_cell, max_dist, min_dist):
        """Take a grid and calculate all grid points with a certain distance cutoff of any point in cart_sites"""

        if min_dist: assert max_dist > min_dist, 'Minimum Mask Distance must be smaller than Maximum Mask Distance'

        # Store distances from masking atoms
        self._max_dist = max_dist
        self._min_dist = min_dist

        # Store grid size
        self._grid_size = grid_size
        self._grid_idxr = flex.grid(grid_size)

        # Unit cell
        self._fake_unit_cell = unit_cell

        # Calculate the masked indices defined by max distance from protein atoms
        self._outer_mask_indices = maptbx.grid_indices_around_sites(unit_cell  = unit_cell,
                                                                    fft_n_real = grid_size, fft_m_real=grid_size,
                                                                    sites_cart = cart_sites,
                                                                    site_radii = flex.double(cart_sites.size(), max_dist))
        outer_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        outer_mask_binary.put(self._outer_mask_indices, True)
        self._outer_mask_binary = outer_mask_binary

        # Calculate the masked indices defined by min distance from protein atoms
        self._inner_mask_indices = maptbx.grid_indices_around_sites(unit_cell  = unit_cell,
                                                                    fft_n_real = grid_size, fft_m_real = grid_size,
                                                                    sites_cart = cart_sites,
                                                                    site_radii = flex.double(cart_sites.size(), min_dist))
        inner_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        inner_mask_binary.put(self._inner_mask_indices, True)
        self._inner_mask_binary = inner_mask_binary

        # Calculate the combination of these masks
        total_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), bool)
        total_mask_binary.put(self._outer_mask_indices, True)
        total_mask_binary.put(self._inner_mask_indices, False)
        self._total_mask_binary = total_mask_binary
        self._total_mask_indices = [idx for idx in xrange(self._grid_idxr.size_1d()) if self._total_mask_binary[idx]>0]

    def total_mask(self):
        """Return the grid points allowed by the mask - combination of max_dist (allowed) and min_dist (rejected)"""
        for p in flex.nested_loop(self._grid_size):
            if self._total_mask_binary[self._grid_idxr(p)]: yield p
    def outer_mask(self):
        """Get grid points allowed subject to max_dist"""
        for p in flex.nested_loop(self._grid_size):
            if self._outer_mask_binary[self._grid_idxr(p)]: yield p
    def inner_mask(self):
        """Get grid points rejected subject to min_dist"""
        for p in flex.nested_loop(self._grid_size):
            if self._inner_mask_binary[self._grid_idxr(p)]: yield p

    def total_mask_binary(self):
        return self._total_mask_binary
    def outer_mask_binary(self):
        return self._outer_mask_binary
    def inner_mask_binary(self):
        return self._inner_mask_binary

    def total_mask_indices(self):
        return self._total_mask_indices
    def outer_mask_indices(self):
        return self._outer_mask_indices
    def inner_mask_indices(self):
        return self._inner_mask_indices

    def total_size(self):
        """Returns the number of grid points in the mask"""
        return len(self._total_mask_indices)
    def outer_size(self):
        """Returns the number of grid points inside max_dist"""
        return len(self._outer_mask_indices)
    def inner_size(self):
        """Returns the number of grid points inside min_dist"""
        return len(self._inner_mask_indices)

    def extent(self):
        """Returns the minimum and maximum grid points in the mask"""
        return min(self.total_mask()), max(self.total_mask())

    def summary(self):
        return '\n'.join([  '----------------------------------->>>',
                            'Atomic Mask Summary:',
                            'Total Mask Size (1D): {!s}'.format(self.total_size()),
                            'Outer Mask Size (1D): {!s}'.format(self.outer_size()),
                            'Inner Mask Size (1D): {!s}'.format(self.inner_size()),
                            'Masked Grid Min/Max: {!s}'.format(self.extent())
                        ])

class grid_mask(atomic_mask):
    """Creates the same mask as `atomic_mask` without allowing for lattice symmetry (rejects atoms outside of the unit cell)"""

    def __init__(self, cart_sites, grid_size, unit_cell, max_dist, min_dist):
        """Mask a grid (approximately) against a set of cartesian points"""

        if min_dist: assert max_dist >= min_dist, 'Minimum Mask Distance must be smaller than Maximum Mask Distance'

        # Store distances from masking atoms
        self._max_dist = max_dist
        self._min_dist = min_dist

        # Store grid size
        self._grid_size = grid_size
        self._grid_idxr = flex.grid(grid_size)

        # Unit cell
        self._fake_unit_cell = unit_cell

        # Filter out sites that are outside of the unit cell
        frac_sites = unit_cell.fractionalize(cart_sites)
        ax1_len_1 = (frac_sites.dot((1.0,0.0,0.0)) <  1.000001).iselection()
        ax2_len_1 = (frac_sites.dot((0.0,1.0,0.0)) <  1.000001).iselection()
        ax3_len_1 = (frac_sites.dot((0.0,0.0,1.0)) <  1.000001).iselection()
        ax1_len_2 = (frac_sites.dot((1.0,0.0,0.0)) > -0.000001).iselection()
        ax2_len_2 = (frac_sites.dot((0.0,1.0,0.0)) > -0.000001).iselection()
        ax3_len_2 = (frac_sites.dot((0.0,0.0,1.0)) > -0.000001).iselection()
        allowed_idxs_1 = ax1_len_1.intersection(ax2_len_1).intersection(ax3_len_1)
        allowed_idxs_2 = ax1_len_2.intersection(ax2_len_2).intersection(ax3_len_2)
        allowed_idxs = allowed_idxs_1.intersection(allowed_idxs_2)
        if not allowed_idxs: raise Exception('No Sites valid for masking!')
        filt_cart_sites = cart_sites.select(allowed_idxs)
#        filt_cart_sites_2 = flex.vec3_double([p for p in cart_sites if not [x for x in unit_cell.fractionalize(p) if (x>1.000001 or x<-0.000001)]])

        # Calculate the masked indices defined by max distance from protein atoms
        self._outer_mask_indices = maptbx.grid_indices_around_sites(unit_cell  = unit_cell,
                                                                    fft_n_real = grid_size, fft_m_real = grid_size,
                                                                    sites_cart = filt_cart_sites,
                                                                    site_radii = flex.double(filt_cart_sites.size(), max_dist))
        outer_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        outer_mask_binary.put(self._outer_mask_indices, True)
        self._outer_mask_binary = outer_mask_binary

        # Calculate the masked indices defined by min distance from protein atoms
        self._inner_mask_indices = maptbx.grid_indices_around_sites(unit_cell  = unit_cell,
                                                                    fft_n_real = grid_size, fft_m_real = grid_size,
                                                                    sites_cart = filt_cart_sites,
                                                                    site_radii = flex.double(filt_cart_sites.size(), min_dist))
        inner_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        inner_mask_binary.put(self._inner_mask_indices, True)
        self._inner_mask_binary = inner_mask_binary

        # Calculate the combination of these masks
        total_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), bool)
        total_mask_binary.put(self._outer_mask_indices, True)
        total_mask_binary.put(self._inner_mask_indices, False)
        self._total_mask_binary  = total_mask_binary
        self._total_mask_indices = [idx for idx in xrange(self._grid_idxr.size_1d()) if self._total_mask_binary[idx]>0]

class non_symmetrical_atomic_mask(atomic_mask):
    """Creates the same mask as `atomic_mask` without allowing for unit cell shifts from the p1 box"""

    def __init__(self, cart_sites, grid_spacing, grid_size, unit_cell, max_dist, min_dist):
        """Take a grid and calculate all grid points with a certain distance cutoff of any point in cart_sites"""

        if min_dist: assert max_dist >= min_dist, 'Minimum Mask Distance must be smaller than Maximum Mask Distance'

        # Store distances from masking atoms
        self._max_dist = max_dist
        self._min_dist = min_dist

        # Store grid size
        self._grid_size = grid_size
        self._grid_idxr = flex.grid(grid_size)

        # Unit cell
        self._fake_unit_cell = unit_cell

        # First, trim off cartesian sites that won't contribute to the masking (those more than max_dist from the edge of the box)
        # Extract the maximum and minimum of the unit cell (assumed to be a p1 grid on the origin)
        uc_min = flex.double([0.0,0.0,0.0])
        uc_max = flex.double(unit_cell.parameters()[:3])
        # We'll be looking for points within max_dist of this grid
        grid_filter_minimum = uc_min - max_dist
        grid_filter_maximum = uc_max + max_dist
        # Filter out sites that are too far from the grid to be worth masking
        filt_cart_sites = flex.vec3_double([p for p in cart_sites if not (    ([i+1 for i in range(3) if p[i]<grid_filter_minimum[i]])
                                                                           or ([i+1 for i in range(3) if p[i]>grid_filter_maximum[i]])    )])

        # Mask the grid
        total_mask_points, \
        outer_mask_points, \
        inner_mask_points = get_grid_points_within_distance_cutoff_of_cart_sites_2( cart_sites=filt_cart_sites,
                                                                                    grid_spacing=grid_spacing,
                                                                                    max_dist=max_dist,
                                                                                    min_dist=min_dist   )

        # Remove points that are outside of the grid we are interested in
        total_mask_points = [gp for gp in total_mask_points if not ( ([i+1 for i in range(3) if gp[i]<=0]) or ([i+1 for i in range(3) if gp[i]>=grid_size[i]]) )]
        outer_mask_points = [gp for gp in outer_mask_points if not ( ([i+1 for i in range(3) if gp[i]<=0]) or ([i+1 for i in range(3) if gp[i]>=grid_size[i]]) )]
        inner_mask_points = [gp for gp in inner_mask_points if not ( ([i+1 for i in range(3) if gp[i]<=0]) or ([i+1 for i in range(3) if gp[i]>=grid_size[i]]) )]

        # Convert the list of points to lists of indices, and binary masks
        self._total_mask_indices = [self._grid_idxr(gp) for gp in total_mask_points]
        total_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        total_mask_binary.put(self._total_mask_indices, True)
        self._total_mask_binary = total_mask_binary

        self._outer_mask_indices = [self._grid_idxr(gp) for gp in outer_mask_points]
        outer_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        outer_mask_binary.put(self._outer_mask_indices, True)
        self._outer_mask_binary = outer_mask_binary

        self._inner_mask_indices = [self._grid_idxr(gp) for gp in inner_mask_points]
        inner_mask_binary = numpy.zeros(self._grid_idxr.size_1d(), dtype=bool)
        inner_mask_binary.put(self._inner_mask_indices, True)
        self._inner_mask_binary = inner_mask_binary

class spherical_mask(object):
    def __init__(self, grid_spacing, distance_cutoff, grid_jump=None):
        """Sphere used to mask grid points within a certain distance of a point"""

        self._mask = get_grid_points_within_distance_cutoff_of_origin(grid_spacing=grid_spacing, distance_cutoff=distance_cutoff)
        self._radius = distance_cutoff
        self._buffer = max(max(self._mask))
        self._grid_spacing = grid_spacing
        if grid_jump:
            self._grid_jump = int(grid_jump)
        else:
            self._grid_jump = iceil(self._buffer)
        if self._grid_jump == 0:
            self._grid_jump = 1

    def mask(self):
        return self._mask
    def size(self):
        return len(self.mask())
    def buffer_size(self):
        return self._buffer
    def grid_spacing(self):
        return self._grid_spacing
    def grid_jump(self):
        return self._grid_jump
    def radius(self):
        return self._radius
    def volume(self):
        return (4/3.0)*numpy.pi*(self.radius()**3)

    def apply_mask(self, grid_point):
        """Combine a grid point with all of the masking vectors"""
        return combine_grid_point_and_grid_vectors(start_point=grid_point, grid_vectors=self.mask())

    def summary(self):
        return '\n'.join(['----------------------------------->>>',
                          'Local Mask Summary:',
                          'Number of Mask Points:  {!s}'.format(len(self.mask())),
                          'Mask Radius (Cart):     {!s}'.format(self.radius()),
                          'Mask Volume (Cart):     {!s}'.format(round(self.volume(),3)),
                          'Largest Mask Vector:    {!s}'.format(max(self.mask())),
                          'Req. Edge Buffer Zone:  {!s}'.format(self.buffer_size()),
                          'Sampling Fraction (1D): 1/{!s}'.format(self.grid_jump()),
                          'Sampling Fraction (3D): 1/{!s}'.format(self.grid_jump()**3)
                        ])





import os, sys, glob, time, gc

#################################
try:
    import matplotlib
    matplotlib.use('Agg')
    matplotlib.interactive(0)
    from matplotlib import pyplot
    pyplot.style.use('ggplot')
except:
    pass
#################################

import numpy

from scitbx.array_family import flex
from scitbx.math import basic_statistics

from libtbx import easy_mp
from libtbx.utils import Sorry, Failure

from bamboo.common import Meta, Info
from bamboo.common.logs import Log
from bamboo.common.path import rel_symlink

from giant.jiffies import extract_params_default
from giant.grid.masks import grid_mask
from giant.xray.maps.bdc import calculate_bdc_subtracted_map, estimate_event_background_correction

from pandda import welcome
from pandda import analyse_html, analyse_graphs
from pandda.phil import pandda_phil
from pandda.analyse_main import PanddaMultiDatasetAnalyser, PanddaMapAnalyser, PanddaZMapAnalyser, MapHolder
from pandda.handlers import DatasetHandler
from pandda.events import PointCluster, Event
from pandda.misc import *

# ============================================================================>
#
###                 Dataset Processing Function
#
# ============================================================================>

def process_dataset_map_func(proc_args):
    """The main function for comparing each dataset-map to the ensemble-maps"""

    # ============================================================================>
    # Unpack input args
    # ============================================================================>
    d_handler              = proc_args[0]
    m_handler              = proc_args[1]
    reference_grid         = proc_args[2]
    reference_dataset      = proc_args[3]
    map_analyser           = proc_args[4]
    args, params, settings = proc_args[5]

    log_strs = []

    # ============================================================================>
    # Build new blob search object
    # ============================================================================>
    print('Analysing Dataset {!s}'.format(d_handler.tag))

    blob_finder = PanddaZMapAnalyser( params       = params.blob_search,
                                      grid_spacing = reference_grid.grid_spacing(),
                                      log          = Log(log_file=d_handler.output_handler.get_file('dataset_log'), verbose=False) )

    # ============================================================================>
    # Generate masks for this dataset
    # ============================================================================>
    print('Generating Symmetry Contacts for Dataset {!s}'.format(d_handler.tag))

    # Generate symmetry contacts for this dataset
    dataset_sym_copies = d_handler.generate_symmetry_copies(rt_method      = params.alignment.method,
                                                            save_operators = False,
                                                            buffer         = params.masks.outer_mask+5   )
    # Extract protein atoms from the symmetry copies
    cache = dataset_sym_copies.atom_selection_cache()
    dataset_sym_sites_cart = dataset_sym_copies.select(cache.selection('pepnames and not element H')).atoms().extract_xyz()
    # Generate custom grid mask for this dataset
    dataset_contact_mask = grid_mask(   cart_sites = dataset_sym_sites_cart,
                                        grid_size  = reference_grid.grid_size(),
                                        unit_cell  = reference_grid.unit_cell(),
                                        max_dist   = params.masks.outer_mask,
                                        min_dist   = params.masks.inner_mask_symmetry )
    # Combine the standard mask with the custom mask
    grid_idxr = reference_grid.grid_indexer()
    dataset_total_mask = [gp for gp in reference_grid.global_mask().total_mask() if dataset_contact_mask.inner_mask_binary()[grid_idxr(gp)] < 0.5]
    dataset_outer_mask = [gp for gp in reference_grid.global_mask().outer_mask() if dataset_contact_mask.inner_mask_binary()[grid_idxr(gp)] < 0.5]
    # Write coordinates of symmetry copies
    dataset_sym_copies.write_pdb_file(d_handler.output_handler.get_file('symmetry_copies'))

    # ============================================================================>
    #####
    # CALCULATE Z-MAPS AND LOOK FOR LARGE BLOBS
    #####
    # ============================================================================>
    print('Calculating Z-MAPs for Dataset {!s}'.format(d_handler.tag))

    # ============================================================================>
    # CALCULATE MEAN-DIFF MAPS
    # ============================================================================>
    assert m_handler.map is not None, 'NO MAP FOUND'
    d_map = m_handler.map - map_analyser.statistical_maps.mean_map

    # ============================================================================>
    # NAIVE Z-MAP - NOT USING UNCERTAINTY ESTIMATION OR ADJUSTED STDS
    # ============================================================================>
    z_map_naive = map_analyser.calculate_z_map(map=m_handler.map, method='naive')
    # Normalise this map to N(0,1)
    z_map_naive_masked     = [z_map_naive[i] for i in reference_grid.global_mask().outer_mask_indices()]
    z_map_naive_normalised = (z_map_naive - numpy.mean(z_map_naive_masked)) / numpy.std(z_map_naive_masked)

    # ============================================================================>
    # UNCERTAINTY Z-MAP - NOT USING ADJUSTED STDS
    # ============================================================================>
    z_map_uncty = map_analyser.calculate_z_map(map=m_handler.map, uncertainty=m_handler.meta.map_uncertainty, method='uncertainty')
    # Normalise this map to N(0,1)
    z_map_uncty_masked     = [z_map_uncty[i] for i in reference_grid.global_mask().outer_mask_indices()]
    z_map_uncty_normalised = (z_map_uncty - numpy.mean(z_map_uncty_masked)) / numpy.std(z_map_uncty_masked)

    # ============================================================================>
    # ADJUSTED+UNCERTAINTY Z-MAP
    # ============================================================================>
    z_map_compl = map_analyser.calculate_z_map(map=m_handler.map, uncertainty=m_handler.meta.map_uncertainty, method='adjusted+uncertainty')
    # Normalise this map to N(0,1)
    z_map_compl_masked     = [z_map_compl[i] for i in reference_grid.global_mask().outer_mask_indices()]
    z_map_compl_normalised = (z_map_compl - numpy.mean(z_map_compl_masked)) / numpy.std(z_map_compl_masked)

    # ============================================================================>
    # ANALYSE Z-MAP FOR STATISTICAL VALIDITY
    # ============================================================================>
    # Calculate statistics of z-maps
    z_map_stats = basic_statistics(flex.double(z_map_compl_masked))
    m_handler.meta.z_mean = z_map_stats.mean
    m_handler.meta.z_stdv = z_map_stats.bias_corrected_standard_deviation
    m_handler.meta.z_skew = z_map_stats.skew
    m_handler.meta.z_kurt = z_map_stats.kurtosis

    # ============================================================================>
    # SELECT WHICH MAP TO DO THE BLOB SEARCHING ON
    # ============================================================================>
    if params.z_map.map_type   == 'naive':
        z_map = z_map_naive_normalised
    elif params.z_map.map_type == 'adjusted':
        z_map = z_map_adjst_normalised
    elif params.z_map.map_type == 'uncertainty':
        z_map = z_map_uncty_normalised
    elif params.z_map.map_type == 'adjusted+uncertainty':
        z_map = z_map_compl_normalised

    # ============================================================================>
    # STORE THE Z MAP
    # ============================================================================>
    z_map_holder = MapHolder(   num         = m_handler.num,
                                tag         = m_handler.tag,
                                map         = z_map,
                                # Change these for the 'fake' grid unit_cell and a P1 space_group
                                unit_cell   = reference_grid.unit_cell(),
                                space_group = reference_grid.space_group(),
                                meta        = Meta({'type'          : 'z-map',
                                                    'resolution'    : m_handler.meta.resolution}),
                                parent      = m_handler  )

    # ============================================================================>
    #####
    # WRITE ALL MAP DISTRIBUTIONS (THESE DON'T USE MUCH SPACE)
    #####
    # ============================================================================>
    # Sampled Map
    write_map_value_distribution(map_vals     = m_handler.map.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('s_map_png'))
    # Mean-Difference
    write_map_value_distribution(map_vals     = d_map.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('d_mean_map_png'))
    # Naive Z-Map
    write_map_value_distribution(map_vals     = z_map_naive.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_naive_png'),
                                 plot_normal  = True)
    # Normalised Naive Z-Map
    write_map_value_distribution(map_vals     = z_map_naive_normalised.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_naive_normalised_png'),
                                 plot_normal  = True)
    # Uncertainty Z-Map
    write_map_value_distribution(map_vals     = z_map_uncty.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_uncertainty_png'),
                                 plot_normal  = True)
    # Normalised Uncertainty Z-Map
    write_map_value_distribution(map_vals     = z_map_uncty_normalised.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_uncertainty_normalised_png'),
                                 plot_normal  = True)
    # Corrected Z-Map
    write_map_value_distribution(map_vals     = z_map_compl.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_corrected_png'),
                                 plot_normal  = True)
    # Normalised Corrected Z-Map
    write_map_value_distribution(map_vals     = z_map_compl_normalised.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_corrected_normalised_png'),
                                 plot_normal  = True)
    # Plot Q-Q Plot of Corrected Z-Map to see how normal it is
    write_qq_plot_against_normal(map_vals     = z_map_compl_normalised.select(reference_grid.global_mask().outer_mask_indices()),
                                 output_file  = d_handler.output_handler.get_file('z_map_qq_plot_png'))

    # ============================================================================>
    #####
    # LOOK FOR CLUSTERS OF LARGE Z-SCORES
    #####
    # ============================================================================>
    # Contour the grid at a particular Z-Value
    # ============================================================================>
    num_clusters, z_clusters = blob_finder.cluster_high_z_values(z_map=z_map, point_mask=dataset_total_mask)
    # ============================================================================>
    # Too many points to cluster -- probably a bad dataset
    # ============================================================================>
    if num_clusters == -1:
        # This dataset is too noisy to analyse - flag!
        #pandda.datasets.all_masks().set_mask_value(mask_name='noisy zmap', entry_id=d_handler.tag, value=True)
        log_strs.append('Z-Map too noisy to analyse')
        return d_handler, m_handler, log_strs

    # ============================================================================>
    #####
    # FILTER/SELECT CLUSTERS OF Z-SCORES
    #####
    # ============================================================================>
    # Filter the clusters by size and peak height
    # ============================================================================>
    if num_clusters > 0:
        num_clusters, z_clusters = blob_finder.filter_z_clusters_1(z_clusters=z_clusters)
        blob_finder.validate_clusters(z_clusters)
        if num_clusters == 0: log_strs.append('===> Minimum cluster peak/size not reached.')
    # ============================================================================>
    # Filter the clusters by distance from protein
    # ============================================================================>
    if num_clusters > 0:
        num_clusters, z_clusters = blob_finder.filter_z_clusters_2(z_clusters       = z_clusters,
                                                                   grid_origin_cart = reference_dataset.origin_shift(),
                                                                   ref_structure    = reference_dataset.new_structure().hierarchy)
        blob_finder.validate_clusters(z_clusters)
        if num_clusters == 0: log_strs.append('===> Clusters too far from protein.')
    # ============================================================================>
    # Group Nearby Clusters Together
    # ============================================================================>
    if num_clusters > 0:
        num_clusters, z_clusters = blob_finder.group_clusters(z_clusters=z_clusters)
        blob_finder.validate_clusters(z_clusters)
    # ============================================================================>
    # Filter the clusters by symmetry equivalence
    # ============================================================================>
    if num_clusters > 0:
        num_clusters, z_clusters = blob_finder.filter_z_clusters_3(z_clusters       = z_clusters,
                                                                   grid_origin_cart = reference_dataset.origin_shift(),
                                                                   ref_unit_cell    = reference_dataset.unit_cell,
                                                                   ref_sym_ops      = reference_dataset.crystal_contact_generators,
                                                                   ref_structure    = reference_dataset.new_structure().hierarchy     )
        blob_finder.validate_clusters(z_clusters)

    # ============================================================================>
    #####
    # WRITE MAPS
    #####
    # ============================================================================>
    if (num_clusters != 0) or args.output.developer.write_maps_for_empty_datasets:
        # ============================================================================>
        # NATIVE MAPS (ROTATED)
        # ============================================================================>
        # Z-map
        write_array_to_map( output_file = d_handler.output_handler.get_file('native_z_map'),
                            map_data    = rotate_map(grid=reference_grid, d_handler=d_handler, map_data=z_map),
                            grid        = reference_grid   )
        # ============================================================================>
        # REFERENCE FRAME MAPS (NOT ROTATED)
        # ============================================================================>
        # Write the sampled map
        write_array_to_map( output_file = d_handler.output_handler.get_file('sampled_map'),
                            map_data    = m_handler.map,
                            grid        = reference_grid   )
        # Write the mean-difference map
        write_array_to_map( output_file = d_handler.output_handler.get_file('mean_diff_map'),
                            map_data    = d_map,
                            grid        = reference_grid   )
        # Write Chosen Z-Map (ALWAYS WRITE THIS MAP)
        write_array_to_map( output_file = d_handler.output_handler.get_file('z_map'),
                            map_data    = z_map,
                            grid        = reference_grid   )
        # ============================================================================>
        # MASKS/BLOBS/GENERIC MAPS
        # ============================================================================>
        if args.output.developer.write_grid_masks:
            # Write map of grid + symmetry mask
            map_mask = numpy.zeros(reference_grid.grid_size_1d(), dtype=int)
            map_mask.put(map(reference_grid.grid_indexer(), dataset_outer_mask), 1)
            map_mask = flex.double(map_mask.tolist()); map_mask.reshape(flex.grid(reference_grid.grid_size()))
            write_array_to_map( output_file = d_handler.output_handler.get_file('grid_mask'),
                                map_data    = map_mask,
                                grid        = reference_grid    )
            # Write map of where the blobs are (high-Z mask)
            highz_points = []; [highz_points.extend(list(x[0])) for x in z_clusters]
            highz_points = [map(int, v) for v in highz_points]
            highz_map_array = numpy.zeros(reference_grid.grid_size_1d(), dtype=int)
            highz_map_array.put(map(reference_grid.grid_indexer(), list(highz_points)), 1)
            map_mask = flex.double(highz_map_array.tolist()); map_mask.reshape(flex.grid(reference_grid.grid_size()))
            write_array_to_map( output_file = d_handler.output_handler.get_file('high_z_mask'),
                                map_data    = map_mask,
                                grid        = reference_grid    )
        # ============================================================================>
        # Write different Z-Maps? (Probably only needed for testing)
        # ============================================================================>
        if args.output.developer.write_all_z_map_types:
            # Write Naive Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_naive'),
                                map_data    = z_map_naive,
                                grid        = reference_grid   )
            # Write Normalised Naive Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_naive_normalised'),
                                map_data    = z_map_naive_normalised,
                                grid        = reference_grid   )
            # Write Uncertainty Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_uncertainty'),
                                map_data    = z_map_uncertainty,
                                grid        = reference_grid   )
            # Write Normalised Uncertainty Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_uncertainty_normalised'),
                                map_data    = z_map_uncertainty_normalised,
                                grid        = reference_grid   )
            # Write Corrected Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_corrected'),
                                map_data    = z_map_compl,
                                grid        = reference_grid   )
            # Write Normalised Corrected Z-Map
            write_array_to_map( output_file = d_handler.output_handler.get_file('z_map_corrected_normalised'),
                                map_data    = z_map_compl_normalised,
                                grid        = reference_grid   )

    # ============================================================================>
    # Skip to next dataset if no clusters found
    # ============================================================================>
    if num_clusters > 0:
        log_strs.append('===> {!s} Cluster(s) found.'.format(num_clusters))
    else:
        log_strs.append('===> No Clusters found.')
        return (d_handler, m_handler, log_strs)
    assert num_clusters > 0, 'NUMBER OF CLUSTERS AFTER FILTERING == 0!'

    # ============================================================================>
    # Process the identified features
    # ============================================================================>
    for event_idx, (event_points, event_values) in enumerate(z_clusters):
        # Number events from 1
        event_num = event_idx + 1
        # Create a unique identifier for this event
        event_key = (d_handler.tag, event_num)
        # ============================================================================>
        # Create a point cluster object
        # ============================================================================>
        point_cluster = PointCluster(id=event_key, points=event_points, values=event_values)
        # ============================================================================>
        # Estimate the background correction of the detected feature
        # ============================================================================>
        # Extract sites for this cluster and estimate the background correction for the event
        log_strs.append('----------------------------------->>>')
        log_strs.append('Estimating Event {!s} Background Correction'.format(event_num))
        # Generate custom grid mask for this dataset
        event_mask = grid_mask( cart_sites = flex.vec3_double(point_cluster.points)*reference_grid.grid_spacing(),
                                grid_size  = reference_grid.grid_size(),
                                unit_cell  = reference_grid.unit_cell(),
                                max_dist   = 1.0,
                                min_dist   = 0.0 )
        expanded_points = list(event_mask.outer_mask())
        log_strs.append('=> Event sites ({!s} points) expanded to {!s} points'.format(len(point_cluster.points), len(expanded_points)))
        # Use the outer mask to provide a reference point for the correlation to the mean map
        reference_points = list(reference_grid.global_mask().inner_mask())
        # Calculate the correlation with the mean map as different amounts of the mean map are subtracted
        event_remain_est = estimate_event_background_correction(
            ref_map          = map_analyser.statistical_maps.mean_map,
            query_map        = m_handler.map,
            feature_region   = expanded_points,
            reference_region = reference_points,
            min_remain       = 1.0-params.background_correction.max_bdc,
            max_remain       = 1.0-params.background_correction.min_bdc,
            bdc_increment    = params.background_correction.increment,
            method           = 'value',
            verbose          = settings.verbose)
        log_strs.append('=> Event Background Correction estimated as {!s}'.format(1-event_remain_est))
        # ============================================================================>
        # Calculate background-corrected map
        # ============================================================================>
        event_map = calculate_bdc_subtracted_map(
            ref_map   = map_analyser.statistical_maps.mean_map,
            query_map = m_handler.map,
            bdc       = 1.0 - event_remain_est)
        # Write out these array (reference and native frames)
        write_array_to_map( output_file = d_handler.output_handler.get_file('event_map').format(event_num, event_remain_est),
                            map_data    = event_map,
                            grid        = reference_grid     )
        write_array_to_map( output_file = d_handler.output_handler.get_file('native_event_map').format(event_num, event_remain_est),
                            map_data    = rotate_map(grid=reference_grid, d_handler=d_handler, map_data=event_map),
                            grid        = reference_grid     )
#                            map_data    = rotate_map(grid=reference_grid, d_handler=d_handler, map_data=event_map, align_on_grid_point=map(int,point_cluster.centroid)))
        # ============================================================================>
        # Find the nearest calpha to the event
        # ============================================================================>
        rt_lab = reference_grid.partition().query_by_grid_points([map(int,point_cluster.centroid)])[0]
        log_strs.append('=> Nearest C-alpha to event: Chain {}, Residue {}'.format(rt_lab[0], rt_lab[1]))
        # ============================================================================>
        # Create an event object
        # ============================================================================>
        event_obj = Event(id=point_cluster.id, cluster=point_cluster)
        event_obj.info.estimated_pseudo_occupancy = event_remain_est
        event_obj.info.estimated_bdc              = 1.0 - event_remain_est
        # ============================================================================>
        # Append to dataset handler
        # ============================================================================>
        d_handler.events.append(event_obj)

        # ============================================================================>
        # Write out the graph of the background correction correlations
        # ============================================================================>
        #if pandda.settings.plot_graphs:
        #    pandda.something()

    return (d_handler, m_handler, log_strs)

# ============================================================================>
#
###                 PanDDA Initialisation Functions
#
# ============================================================================>

def pandda_dataset_setup(pandda):
    """Initialise the pandda object and load the input datasets"""

    pandda.log('',True)
    pandda.log('=====================================================>>>', True)
    pandda.log('Running PanDDA Setup', True)
    pandda.log('=====================================================>>>', True)
    pandda.log('',True)

    # ============================================================================>
    #####
    # Build list of files in data directories
    #####
    # ============================================================================>
    input_files = pandda.build_input_list()
    # Check that some datasets have been found or already loaded
    if (not pandda.datasets.all()) and (not input_files):
        raise Sorry('NO DATASETS HAVE BEEN SELECTED FOR ANALYSIS OR LOADED FROM PREVIOUS RUNS')
    # Check to see if we're reusing statistical maps
    if (not pandda.args.method.recalculate_statistical_maps) and pandda.stat_maps.get_resolutions():
        pandda.log('----------------------------------->>>', True)
        pandda.log('Pre-existing statistical maps (from previous runs) have been found and will be reused:', True)
        pandda.log('Resolutions of reloaded maps: {!s}'.format(', '.join(map(str,pandda.stat_maps.get_resolutions()))), True)
    # Check that enough datasets have been found
    elif pandda.datasets.size() + len(input_files) < pandda.params.analysis.min_build_datasets:
        pandda.log('----------------------------------->>>', True)
        pandda.log('NOT ENOUGH DATASETS HAVE BEEN LOADED FOR ANALYSIS', True)
        pandda.log('Number loaded ({!s}) is less than the {!s} needed.'.format(pandda.datasets.size()+len(input_files), pandda.params.analysis.min_build_datasets), True)
        pandda.log('This value is controlled by changing pandda.params.analysis.min_build_datasets', True)
        raise Sorry('NOT ENOUGH DATASETS HAVE BEEN LOADED FOR ANALYSIS')
    # If dry_run, exit after initial search
    if pandda.args.exit_flags.dry_run:
        raise SystemExit('Dry Run Only: Exiting')
    # Load and process input files
    if input_files and (pandda.args.input.max_new_datasets > 0):
        # ============================================================================>
        # Limit the number of files that can be added at once to a pandda
        # ============================================================================>
        if len(input_files) > pandda.args.input.max_new_datasets:
            pandda.log('----------------------------------->>>', True)
            pandda.log('Limiting the number of new datasets that are added to the pandda analysis (controlled by input.max_new_datasets)')
            pandda.log('Limiting analysis to the first {} of {} datasets'.format(pandda.args.input.max_new_datasets, len(input_files)))
            input_files = input_files[:pandda.args.input.max_new_datasets]
            assert len(input_files) == pandda.args.input.max_new_datasets, 'Something has gone wrong'
        # ============================================================================>
        #####
        # Add new files and load datasets
        #####
        # ============================================================================>
        pandda.add_new_files(input_files)
        pandda.load_new_datasets()
        pandda.initialise_dataset_masks_and_tables()
        pandda.check_loaded_datasets(datasets=pandda.datasets.all())
        # ============================================================================>
        #####
        # Set Reference Dataset
        #####
        # Select the reference dataset
        # ============================================================================>
        if not pandda.reference_dataset():
            # Filter datasets against the provided filter pdb if given
            if pandda.args.input.filter.pdb is not None:
                pandda.log('----------------------------------->>>', True)
                pandda.log('Filtering datasets against the provided pdb structure (defined by pandda.input.filter.pdb)', True)
                pandda.filter_datasets_1(filter_dataset=DatasetHandler(dataset_number=0, pdb_filename=pandda.args.input.filter.pdb))
            # Use given reference dataset, or select reference dataset
            if pandda.args.input.reference.pdb and pandda.args.input.reference.mtz:
                pandda.log('----------------------------------->>>', True)
                pandda.log('Reference Provided by User', True)
                ref_pdb, ref_mtz = pandda.args.input.reference.pdb, pandda.args.input.reference.mtz
            else:
                pandda.log('----------------------------------->>>', True)
                pandda.log('Selecting reference dataset from loaded datasets', True)
                ref_pdb, ref_mtz = pandda.select_reference_dataset(method='resolution')
            # Load the reference dataset
            pandda.load_reference_dataset(ref_pdb=ref_pdb, ref_mtz=ref_mtz)
        # ============================================================================>
        #####
        # Scale, Align and Initial-Filter All Data
        #####
        # ============================================================================>
        # Filter out datasets with different protein structures
        pandda.filter_datasets_1()
        # Load reflection data for each dataset
        pandda.load_reflection_data(ampl_label=pandda.params.maps.ampl_label,
                                    phas_label=pandda.params.maps.phas_label)
        # Align structures to the reference
        pandda.align_datasets(method=pandda.params.alignment.method)
        # Pickle the new loaded datasets
        pandda.pickle_the_pandda(components=['datasets'], datasets=pandda.datasets.mask(mask_name='old datasets', invert=True))
    else:
        # ============================================================================>
        # Rebuild the masks of the rejected datasets (quick)
        # ============================================================================>
        pandda.initialise_dataset_masks_and_tables()
        pandda.check_loaded_datasets(datasets=pandda.datasets.all())
        pandda.filter_datasets_1()

    # ============================================================================>
    # Check that enough VALID datasets have been found
    # ============================================================================>
    if (not pandda.args.method.recalculate_statistical_maps) and pandda.stat_maps.get_resolutions():
        # Using existing maps - don't need to check
        pass
    elif pandda.datasets.size(mask_name='rejected - total', invert=True) < pandda.params.analysis.min_build_datasets:
        pandda.log('----------------------------------->>>', True)
        pandda.log('NOT ENOUGH (NON-REJECTED) DATASETS TO BUILD DISTRIBUTIONS!', True)
        pandda.log('Number loaded ({!s}) is less than the {!s} needed.'.format(pandda.datasets.size(mask_name='rejected - total', invert=True), pandda.params.analysis.min_build_datasets), True)
        pandda.log('This value is defined by pandda.params.analysis.min_build_datasets', True)
        raise Sorry('NOT ENOUGH DATASETS LOADED')

    # ============================================================================>
    #####
    # Filter and Analyse the Datasets
    #####
    # ============================================================================>
    # Collate dataset variables and parameters
    pandda.collate_dataset_variables()
    # Filter out the datasets that are not isomorphous and therefore incomparable
    pandda.filter_datasets_2()
    # Calculate the calpha rmsd to the reference dataset
    pandda.calculate_dataset_rmsds_to_reference()
    # Calculate the variation between the alignments of neighbouring C-alphas
    pandda.analyse_alignment_variation()

    return

def pandda_grid_setup(pandda):
    """Generate the grid objects for the analysis"""

    pandda.log('',True)
    pandda.log('=====================================================>>>', True)
    pandda.log('Running Grid Setup',True)
    pandda.log('=====================================================>>>', True)
    pandda.log('',True)

    # ============================================================================>
    # Re-select the reference dataset - NOT YET IMPLEMENTED (TODO)
    # ============================================================================>
#    highest_res_dh = sorted(pandda.datasets.mask(mask_name='rejected - total', invert=True), key=lambda d: d.reflection_data().max_min_resolution()[1])[0]
#    pandda.log('Reselected Reference Dataset (used for map scaling): {}'.format(highest_res_dh.tag))

    # ============================================================================>
    #####
    # Create Sampling Grid (for generated maps)
    #####
    # Create reference grid based on the reference structure
    # ============================================================================>
    if pandda.reference_grid() is None:
        # Create parameter for setting grid spacing (multiple grids?)
        pandda.create_reference_grid(   grid_spacing     = pandda.params.maps.grid_spacing,
                                        expand_to_origin = False,
                                        buffer           = pandda.params.masks.outer_mask+pandda.params.maps.padding    )
        # Create various masks to define regions of the grid by distance to the protein and symmetry copies
        pandda.mask_reference_grid()
    # ============================================================================>
    # Store for reuse
    # ============================================================================>
    # Pickle all of the large arrays so they can be reloaded
    # ============================================================================>
    pandda.pickle_the_pandda(components=['grid'])

    # ============================================================================>
    # If setup_only, exit after initial search
    # ============================================================================>
    if pandda.args.exit_flags.setup_only:
        pandda.exit(error=False)
        raise SystemExit('Setup Only: Exiting')

    return

# ============================================================================>
#
###                 PanDDA Processing Functions
#
# ============================================================================>

def pandda_main_loop(pandda):
    """Calculate the statistical maps, and then look for events in each dataset"""

    pandda.log('',True)
    pandda.log('=====================================================>>>', True)
    pandda.log('Running Main PanDDA Analysis', True)
    pandda.log('=====================================================>>>', True)
    pandda.log('',True)

    # ============================================================================>
    # Validate/Reset the loaded datasets (those to be analysed)
    # ============================================================================>
    pandda.reset_loaded_datasets()

    # ============================================================================>
    #####
    # PRE-ANALYSIS ANALYSIS (DUMP OF DATASET PARAMETERS)
    #####
    # ============================================================================>
    pandda.write_output_csvs()
    analyse_graphs.write_dataset_summary_graphs(pandda)
    analyse_html.write_initial_html(pandda)

    # ============================================================================>
    #####
    # Update Settings
    #####
    # ============================================================================>
    # Update the resolution limits using the resolution limits from the datasets supplied
    if pandda.params.analysis.dynamic_res_limits:
        pandda.log('----------------------------------->>>')
        pandda.log('UPDATING RESOLUTION LIMITS -')
        pandda.set_low_resolution(  min(pandda.params.analysis.high_res_lower_limit,
                                    max(pandda.tables.dataset_info['high_resolution'])))
        pandda.set_high_resolution( max(pandda.params.analysis.high_res_upper_limit,
                                    pandda.reference_dataset().mtz_summary.high_res,
                                    min(pandda.tables.dataset_info['high_resolution'])))
    else:
        pandda.log('----------------------------------->>>')
        pandda.log('**NOT** UPDATING RESOLUTION LIMITS -')
        pandda.set_low_resolution(  pandda.params.analysis.high_res_lower_limit)
        pandda.set_high_resolution( pandda.params.analysis.high_res_upper_limit)
    # Print new resolution limits
    pandda.log('LOW RESOLUTION:  {!s}'.format(pandda.get_low_resolution()))
    pandda.log('HIGH RESOLUTION: {!s}'.format(pandda.get_high_resolution()))

    # ============================================================================>
    #####
    # PREPARE VARIABLES TO LOOP OVER RESOLUTION SHELLS
    #####
    # ============================================================================>
    # Calculate cutoffs for resolution shells
    # ============================================================================>
    if (not pandda.args.method.recalculate_statistical_maps) and pandda.stat_maps.get_resolutions():
        # Use the resolution limits of previously run campaigns
        res_limits = pandda.stat_maps.get_resolutions()
        assert res_limits, 'No Resolution Limits found from statistical maps: {!s}'.format(res_limits)
        min_limit = min(res_limits)
        max_limit = max(res_limits)
    else:
        # Set pandda.args.method.recalculate_statistical_maps to True as no statistical maps have been found
        if not pandda.args.method.recalculate_statistical_maps:
            pandda.log('No Statistical Maps Found: Setting pandda.args.method.recalculate_statistical_maps to True', True)
            pandda.args.method.recalculate_statistical_maps = True
        # ============================================================================>
        # Select resolution limits based on dataset resolutions and given arguments
        # ============================================================================>
        if pandda.params.analysis.dynamic_res_limits:
            # Round the high limit DOWN to the nearest 0.01
            min_limit = round(pandda.get_high_resolution()- 0.005, 2)    # i.e. 1.344 -> 1.34
            # Round the low limit UP to the nearest 0.01
            max_limit = round(pandda.get_low_resolution() + 0.005, 2)    # i.e. 3.423 -> 3.43
        else:
            # Take the arguments as given by the user
            min_limit = pandda.get_high_resolution()
            max_limit = pandda.get_low_resolution()
        # ============================================================================>
        # Create resolution shells (or single limit)
        # ============================================================================>
        if not pandda.params.analysis.high_res_increment:
            # No variable cutoff - select all
            res_limits = [max_limit]  # i.e. [2]
        else:
            # Calculate a range of resolution limits
            shell_width = pandda.params.analysis.high_res_increment
            res_limits = [round(x, 4) for x in numpy.arange(min_limit, max_limit, shell_width).tolist()]
            # Append the rounded max_limit to the end to ensure that the last dataset is processed
            if (not res_limits) or (res_limits[-1] != max_limit):
                res_limits.append(max_limit)
    # ==================================================>
    #####
    # Initialise for iterations over shells
    #####
    # ============================================================================>
    # Analyse all datasets from high_shell_limit -> cut_resolution (initialise high_shell_limit to 0)
    high_shell_limit = 0
    # Record how many datasets are processed at each resolution
    resolution_count = {}
    # ============================================================================>
    # Report
    # ============================================================================>
    pandda.log('----------------------------------->>>', True)
    if len(res_limits)==1:
        pandda.log('Analysing All Maps at {!s}A'.format(max_limit), True)
    else:
        pandda.log('Analysing Resolution Shells from {!s} -> {!s}A'.format(min_limit, max_limit), True)
        pandda.log('Limits: {!s}'.format(', '.join(map(str,res_limits))), True)

    # ============================================================================>
    #####
    # ANALYSE DATASETS - ITERATE THROUGH RESOLUTION SHELLS
    #####
    # ============================================================================>
    t_analysis_start = time.time()
    # ==================================================>
    pandda.log('----------------------------------->>>', True)
    pandda.log('Dataset Analysis Started: {!s}'.format(time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime(t_analysis_start))), True)

    for cut_resolution in res_limits:

        # Which resolutions will be processed in this shell
        pandda.log('', True)
        pandda.log('=====================================================>>>', True)
        pandda.log('Processing Datasets from {!s}A -> {!s}A'.format(high_shell_limit, cut_resolution), True)
        pandda.log('=====================================================>>>', True)
        pandda.log('',True)

        # ============================================================================>
        # Select datasets to build the distributions
        # ============================================================================>
        building_mask_name = 'selected for building @ {!s}A'.format(cut_resolution)
        if pandda.args.method.recalculate_statistical_maps:
            pandda.log('----------------------------------->>>', True)
            pandda.log('Selecting Building Mask', True)
            building_mask = pandda.select_for_building_distributions(high_res_cutoff = cut_resolution,
                                                                     building_mask_name = building_mask_name)
            # Check that we have enough datasets to build distributions
            if sum(building_mask) < pandda.params.analysis.min_build_datasets:
                # Don't have enough to construct robust distributions
                pandda.log('NOT ENOUGH DATASETS TO CALCULATE DISTRIBUTIONS ({!s}<{!s})'.format(sum(building_mask),pandda.params.analysis.min_build_datasets), True)
                continue
            else:
                # Have enough to generate robust distributions
                pandda.log('ENOUGH DATASETS -> PROCESSING THIS RESOLUTION', True)
                pandda.log('Building Distributions using {!s} Datasets'.format(sum(building_mask)), True)
        else:
            pandda.log('----------------------------------->>>', True)
            pandda.log('**NOT** Selecting Building Mask (Using Existing Statistical Maps)', True)
            # Create a dummy mask as we won't be using any datasets for building
            building_mask = [False]*pandda.datasets.size()
            pandda.datasets.all_masks().add_mask(mask_name=building_mask_name, mask=building_mask)

        # ============================================================================>
        # Select the datasets to analyse
        # ============================================================================>
        analysis_mask_name = 'selected for analysis @ {!s}A'.format(cut_resolution)
        pandda.log('----------------------------------->>>', True)
        pandda.log('Selecting Analysis Mask', True)
        analysis_mask = pandda.select_for_analysis(high_res_large_cutoff = cut_resolution,
                                                   high_res_small_cutoff = high_shell_limit,
                                                   analysis_mask_name = analysis_mask_name)
        # Check that there're some datasets to analyse
        if sum(analysis_mask) == 0:
            pandda.log('NO DATASETS TO ANALYSE @ {!s}'.format(cut_resolution))
            continue
        else:
            pandda.log('Calculating Z-Maps for {!s} Datasets'.format(sum(analysis_mask)), True)

        # ============================================================================>
        # Combine the masks as we will need to load maps for all datasets
        # ============================================================================>
        pandda.log('----------------------------------->>>', True)
        pandda.log('Combining (Analysis and Building) Masks', True)
        map_load_mask = pandda.datasets.all_masks().combine_masks([analysis_mask_name, building_mask_name])
        map_load_mask_name = 'selected for loading maps @ {!s}A'.format(cut_resolution)
        pandda.datasets.all_masks().add_mask(mask_name=map_load_mask_name, mask=map_load_mask)

        # ============================================================================>
        # Report
        # ============================================================================>
        pandda.log('----------------------------------->>>')
        pandda.log('Mask Names for Building, Loading, and Analysis')
        pandda.log('Building ({!s} datasets): {!s}'.format(sum(building_mask), building_mask_name))
        pandda.log('Load Map ({!s} datasets): {!s}'.format(sum(map_load_mask), map_load_mask_name))
        pandda.log('Analysis ({!s} datasets): {!s}'.format(sum(analysis_mask), analysis_mask_name))
        pandda.log('----------------------------------->>>', True)
        pandda.log('Loading Maps for {!s} Datasets at {!s}A'.format(pandda.datasets.size(mask_name=map_load_mask_name), cut_resolution), True)

        # ============================================================================>
        #####
        # LOAD AND ANALYSE MAPS
        #####
        # ============================================================================>
        t_loop_start = time.time()
        # ==================================================>
        pandda.log('{!s}A Analysis Started: {!s}'.format(cut_resolution, time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime(t_loop_start))), True)

        # ============================================================================>
        # Update limit (cut resolution becomes top limit in next shell)
        # ============================================================================>
        high_shell_limit = cut_resolution

        # ============================================================================>
        #####
        # Truncate data, load and scale maps to reference dataset
        #####
        # ============================================================================>
        # Truncate the data to a particular resolution
        # ============================================================================>
        pandda.truncate_scaled_data(dataset_handlers = pandda.datasets.mask(mask_name=map_load_mask_name))
        # ============================================================================>
        # Load the reference map so that we can scale the individual maps to this
        # ============================================================================>
        ref_map_holder = pandda.load_reference_map( map_resolution = cut_resolution )
        # ============================================================================>
        # Load the required maps
        # ============================================================================>
        map_holder_list = pandda.load_and_morph_maps( dataset_handlers = pandda.datasets.mask(mask_name=map_load_mask_name),
                                                      ref_map_holder   = ref_map_holder,
                                                      map_resolution   = cut_resolution     )
        # ============================================================================>
        # Extract the statistical maps at this resolution (if requested)
        # ============================================================================>
        if pandda.args.method.recalculate_statistical_maps:
            # No Statistical maps - set to None so that they are created
            statistical_maps = None
        else:
            # Try to find a statistical map at this resolution
            statistical_maps = pandda.stat_maps.get(cut_resolution)
        # ============================================================================>
        # Create an object to hold all of the maps, and can be used to calculate the mean maps, etc...
        # ============================================================================>
        map_analyser = PanddaMapAnalyser(   dataset_maps     = map_holder_list,
                                            meta             = Meta({'resolution'   : cut_resolution,
                                                                     'grid_size'    : pandda.reference_grid().grid_size(),
                                                                     'grid_size_1d' : pandda.reference_grid().grid_size_1d()}),
                                            statistical_maps = statistical_maps,
                                            parent           = pandda,
                                            log              = pandda.log   )
        # ============================================================================>
        # Add the analysis mask to the map analyser
        # ============================================================================>
        map_analyser.dataset_maps.all_masks().add_mask(mask_name=analysis_mask_name, mask=[False]*map_analyser.dataset_maps.size())
        for dh in pandda.datasets.mask(mask_name=analysis_mask_name):
            map_analyser.dataset_maps.all_masks().set_mask_value(mask_name=analysis_mask_name, entry_id=dh.tag, value=True)
        # ============================================================================>
        # Add the building mask to the map analyser
        # ============================================================================>
        map_analyser.dataset_maps.all_masks().add_mask(mask_name=building_mask_name, mask=[False]*map_analyser.dataset_maps.size())
        for dh in pandda.datasets.mask(mask_name=building_mask_name):
            map_analyser.dataset_maps.all_masks().set_mask_value(mask_name=building_mask_name, entry_id=dh.tag, value=True)

        # ============================================================================>
        #####
        # Calculate Statistical Maps (if required) and Calculate Dataset Uncertainties
        #####
        # ============================================================================>
        if pandda.args.method.recalculate_statistical_maps:
            pandda.log('----------------------------------->>>', True)
            pandda.log('Building Map Distributions for {!s} Datasets at {!s}A'.format(pandda.datasets.size(mask_name=building_mask_name), cut_resolution), True)
        else:
            pandda.log('----------------------------------->>>', True)
            pandda.log('Using Existing Map Distributions at {!s}A'.format(cut_resolution), True)
            assert pandda.datasets.size(mask_name=building_mask_name) == 0, 'BUILDING MASKS HAVE BEEN SELECTED WHEN MAPS ALREADY EXIST'
            assert map_analyser.dataset_maps.size(mask_name=building_mask_name) == 0, 'BUILDING MASKS HAVE BEEN SELECTED WHEN MAPS ALREADY EXIST'
        # ============================================================================>
        # Calculate the mean map
        # ============================================================================>
        if (not map_analyser.statistical_maps.mean_map) or pandda.args.method.recalculate_statistical_maps:
            map_analyser.calculate_mean_map(masked_idxs = pandda.reference_grid().global_mask().outer_mask_indices(),
                                            mask_name   = building_mask_name)
        # ============================================================================>
        # If only Mean Map Requested -- exit
        # ============================================================================>
        if pandda.args.exit_flags.calculate_first_mean_map_only:
            write_array_to_map( output_file = pandda.output_handler.get_file('mean_map').format(cut_resolution),
                                map_data    = map_analyser.statistical_maps.mean_map,
                                grid        = pandda.reference_grid()     )
            raise SystemExit('Calculating First Mean Map Only: Exiting')
        # ============================================================================>
        # Plot Mean Map against Reference Map - should be fairly similar...
        # ============================================================================>
        # Plot the mean map against the reference map (unsorted)
        analyse_graphs.mean_obs_scatter(        f_name    = os.path.join(pandda.output_handler.get_dir('reference'), 'reference_against_mean_unsorted.png'),
                                                mean_vals = map_analyser.statistical_maps.mean_map.select(pandda.reference_grid().global_mask().outer_mask_indices()),
                                                obs_vals  = ref_map_holder.map.select(pandda.reference_grid().global_mask().outer_mask_indices())           )
        # Plot the mean map against the reference map (sorted)
        analyse_graphs.sorted_mean_obs_scatter( f_name    = os.path.join(pandda.output_handler.get_dir('reference'), 'reference_against_mean_sorted.png'),
                                                mean_vals = sorted(map_analyser.statistical_maps.mean_map.select(pandda.reference_grid().global_mask().outer_mask_indices())),
                                                obs_vals  = sorted(ref_map_holder.map.select(pandda.reference_grid().global_mask().outer_mask_indices()))   )
        # Plot the reference map distribution
        write_map_value_distribution( map_vals    = ref_map_holder.map.select(pandda.reference_grid().global_mask().outer_mask_indices()),
                                      output_file = os.path.join(pandda.output_handler.get_dir('reference'), 'reference_map_distribution.png')     )
        # ============================================================================>
        # Calculate the uncertainty of all loaded maps (needs the mean map to have been calculated)
        # ============================================================================>
        # If not pandda.args.method.recalculate_statistical_maps, then no building_mask datasets will have been added, so don't need to be selective in this function
        map_analyser.calculate_map_uncertainties(masked_idxs=pandda.reference_grid().global_mask().inner_mask_indices())
        # ============================================================================>
        # Plot uncertainties of maps
        # ============================================================================>
        try:
            from ascii_graph import Pyasciigraph
            g=Pyasciigraph()
            graph_data = [(mh.tag, round(mh.meta.map_uncertainty,3)) for mh in map_analyser.dataset_maps.all()]
            pandda.log('----------------------------------->>>', True)
            for l in g.graph(label='Sorted Map Uncertainties (Ascending Order)', data=graph_data, sort=1):
                if l.startswith('#######'): continue
                pandda.log(l.replace(u"\u2588", '=').replace('= ','> '), True)
            pandda.log('----------------------------------->>>', True)
            pandda.log('')
        except ImportError:
            print('IMPORT ERROR (ascii_graph) - CANNOT GENERATE UNCERTAINTY GRAPH')
        except:
            pass
        # ============================================================================>
        # Calculate the statistics of the maps
        # ============================================================================>
        if (not map_analyser.statistical_maps.sadj_map) or pandda.args.method.recalculate_statistical_maps:
            map_analyser.calculate_statistical_maps(masked_idxs = pandda.reference_grid().global_mask().outer_mask_indices(),
                                                    mask_name   = building_mask_name,
                                                    cpus        = pandda.settings.cpus)

#        # ============================================================================>
#        # PICKLE THE DATASETS THAT HAVE JUST BEEN PROCESSED
#        # ============================================================================>
#        pandda.sync_datasets(datasets=pandda.datasets.mask(mask_name=analysis_mask_name), overwrite_dataset_meta=True)
#        pandda.pickle_the_pandda(components=['datasets'], datasets=pandda.datasets.mask(mask_name=analysis_mask_name))
        # ============================================================================>
        # Extract and store the statistical map objects
        # ============================================================================>
        # Only need to add if we're calculating new statistical maps
        if pandda.args.method.recalculate_statistical_maps:
            # Check to see if already in there
            if cut_resolution in pandda.stat_maps.get_resolutions():
                pandda.log('Overwriting existing statistical maps @ {!s}A'.format(cut_resolution))
            pandda.stat_maps.add(stat_map_list=map_analyser.statistical_maps, resolution=cut_resolution, overwrite=True)
            pandda.pickle_the_pandda(components=['stat_maps'])
        # ============================================================================>
        # Pickle map analyser if required
        # ============================================================================>
        if pandda.args.output.pickling.pickle_map_analysers:
            pandda.pickle(pickle_file=pandda.pickle_handler.get_file('map_analyser').format(cut_resolution), pickle_object=map_analyser, overwrite=True)
        # ============================================================================>
        # Write Grid Point Distributions - Standard Function, just to provide an output
        # ============================================================================>
        try:
            from libtbx.math_utils import iceil
            grid_size = pandda.reference_grid().grid_size()
            num_points = min(10, min(grid_size))
            grid_points = zip(*[range(0, s, iceil(s/num_points)) for s in grid_size])
            pandda.write_grid_point_distributions(  grid_points     = grid_points,
                                                    map_analyser    = map_analyser,
                                                    output_filename = None          )
        except:
            print('UNIMPORTANT: FAILED TO WRITE AUTOMATIC DISTRIBUTION OF GRID POINTS')
            raise
        # ============================================================================>
        # Write out Grid Point Distributions for interesting grid points (high modality, etc...)
        # ============================================================================>
        # TODO TODO TODO
        # TODO TODO TODO

        # ============================================================================>
        #####
        # DATA PROCESSING
        #####
        # ============================================================================>
        assert cut_resolution not in resolution_count
        resolution_count[cut_resolution] = []
        # ============================================================================>
        # Create a new "dummy" map analyser for the parallel steps
        # ============================================================================>
        dummy_map_analyser = PanddaMapAnalyser( dataset_maps = None,
                                                meta         = Meta({'resolution'   : cut_resolution,
                                                                     'grid_size'    : pandda.reference_grid().grid_size(),
                                                                     'grid_size_1d' : pandda.reference_grid().grid_size_1d()}),
                                                statistical_maps = map_analyser.statistical_maps,
                                                parent           = None,
                                                log              = None   )
        # ============================================================================>
        # Blob Search Object
        # ============================================================================>
        blob_finder = PanddaZMapAnalyser( params = pandda.params.blob_search,
                                          grid_spacing = pandda.reference_grid().grid_spacing(),
                                          log = pandda.log )
        blob_finder.print_settings()

        # ============================================================================>
        #####
        # Calculate Z-Maps
        #####
        # ============================================================================>
        t_anal_start = time.time()
        # ==================================================>
        pandda.log('----------------------------------->>>', True)
        pandda.log('Preparing to Analyse {!s} Datasets at {!s}A'.format(pandda.datasets.size(mask_name=analysis_mask_name), cut_resolution), True)

        proc_dataset_inp_list = []

        # Iterate through and prepare to calculate z-maps
        for i_dh, d_handler in enumerate(pandda.datasets.mask(mask_name=analysis_mask_name)):

            # Record the which resolution this dataset was analysed at
            resolution_count[cut_resolution].append(d_handler.tag)

            # ============================================================================>
            # Validate/Check/Zero the dataset
            # ============================================================================>
            # Dataset should not have any events
            assert d_handler.events == []

            # ============================================================================>
            # Update datasets masks flag - for this analysis
            # ============================================================================>
            pandda.datasets.all_masks().set_mask_value(mask_name='analysed', entry_id=d_handler.tag, value=True)

            # ============================================================================>
            # Update the dataset meta object -- this is persistent
            # ============================================================================>
            d_handler.meta.analysed = True

            # ============================================================================>
            # Extract the map holder for this dataset
            # ============================================================================>
            m_handler = map_analyser.dataset_maps.get(tag=d_handler.tag)

            # ============================================================================>
            # Compile arguments for this datasets
            # ============================================================================>
            dataset_args = (    d_handler,
                                m_handler,
                                pandda.reference_grid(),
                                pandda.reference_dataset(),
                                dummy_map_analyser,
                                (pandda.args, pandda.params, pandda.settings)   )

            proc_dataset_inp_list.append(dataset_args)

#                # Input for process function
#                # Need to create map_analyser for each dataset (can fill with only stat maps to calculate Z-map)
#                d_handler = args[0]
#                m_handler = args[1]
#                ref_grid  = args[2]
#                map_alysr = args[3]
#                params    = args[4]

            ### XXX ------- preprocessing for the map func above here ------- XXX ###

        pandda.log('----------------------------------->>>', True)
        pandda.log('Calculating and Analysing Z-Maps for {!s} Dataset(s) at {!s}A'.format(pandda.datasets.size(mask_name=analysis_mask_name), cut_resolution), True)

        proc_results = easy_mp.pool_map(func=process_dataset_map_func, args=proc_dataset_inp_list,
                                        processes=pandda.settings.cpus, maxtasksperchild=1)

        pandda.log('----------------------------------->>>', True)
        pandda.log('Updating with results from analysing {!s} Dataset(s) at {!s}A'.format(pandda.datasets.size(mask_name=analysis_mask_name), cut_resolution), True)

        for results in proc_results:

            # Unpack results
            d_handler, m_handler, log_strs = results

            pandda.log('', True)
            pandda.log('======================================>>>', True)
            pandda.log('Z-Map Analysis Results for {}'.format(d_handler.tag), True)
            pandda.log('======================================>>>', True)
            pandda.log('\n'.join(log_strs), True)

            # ============================================================================>
            # STORE ANALYSIS DATA IN DATASET MAP TABLE
            # ============================================================================>
            # Add to the dataset map summary table
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'analysed_resolution', m_handler.meta.resolution)
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'map_uncertainty',     round(m_handler.meta.map_uncertainty,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'obs_map_mean',        round(m_handler.meta.obs_map_mean,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'obs_map_rms',         round(m_handler.meta.obs_map_rms,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'z_map_mean',          round(m_handler.meta.z_mean,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'z_map_std',           round(m_handler.meta.z_stdv,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'z_map_skew',          round(m_handler.meta.z_skew,3))
            pandda.tables.dataset_map_info.set_value(d_handler.tag, 'z_map_kurt',          round(m_handler.meta.z_kurt,3))

            # ============================================================================>
            # WRITE OUT DATASET INFORMATION TO CSV FILE
            # ============================================================================>
            out_list = pandda.tables.dataset_info.loc[d_handler.tag].append(pandda.tables.dataset_map_info.loc[d_handler.tag])
            out_list.to_csv(path=d_handler.output_handler.get_file('dataset_info'), header=True, index_label='dtag')

            # ============================================================================>
            # Link outputs for datasets with hits
            # ============================================================================>
            if d_handler.events:
                # Create a link to the interesting directories in the initial results directory
                hit_dir = os.path.join(pandda.output_handler.get_dir('interesting_datasets'), d_handler.tag)
                if not os.path.exists(hit_dir): rel_symlink(orig=d_handler.output_handler.get_dir('root'), link=hit_dir)
                # ============================================================================>
                # Add event to the event table
                # ============================================================================>
                for e in d_handler.events:
                    pandda.add_event_to_event_table(d_handler=d_handler, event=e)

            # ============================================================================>
            # Update the master copy of the dataset object
            # ============================================================================>
            master_d_handler = pandda.datasets.get(tag=d_handler.tag)
            master_d_handler.events = d_handler.events
            master_d_handler.child  = m_handler

        # ============================================================================>
        #####
        # Write the summaries
        #####
        # ============================================================================>
        # Write statistical maps
        pandda.write_map_analyser_maps(map_analyser=map_analyser, analysis_mask_name=analysis_mask_name)
        # Write output graphs for map analysis
        if pandda.settings.plot_graphs:
            analyse_graphs.write_map_analyser_graphs(pandda=pandda, map_analyser=map_analyser, analysis_mask_name=analysis_mask_name)

        # ============================================================================>
        #####
        # Print summaries
        #####
        # ============================================================================>
        t_loop_end = time.time()
        # ==================================================>
        pandda.log('')
        pandda.log('=====================================================>>>', True)
        pandda.log('{!s}A Z-Map Processing Time: {!s}'.format(cut_resolution, time.strftime("%H hours:%M minutes:%S seconds", time.gmtime(t_loop_end - t_anal_start))), True)
        pandda.log('{!s}A Total Processing Time: {!s}'.format(cut_resolution, time.strftime("%H hours:%M minutes:%S seconds", time.gmtime(t_loop_end - t_loop_start))), True)
        pandda.log('=====================================================>>>', True)
        pandda.log('@{!s}A:\t {!s}/{!s} New Datasets Analysed'.format(cut_resolution, pandda.datasets.size(mask_name=analysis_mask_name), pandda.datasets.size(mask_name='old datasets', invert=True)))
        pandda.log('Total:\t {!s}/{!s} New Datasets Analysed'.format(pandda.datasets.size(mask_name='analysed'), pandda.datasets.size(mask_name='old datasets', invert=True)))
        pandda.log('=====================================================>>>', True)
        pandda.log('')

        # ============================================================================>
        #####
        # PICKLE THE DATASETS THAT HAVE JUST BEEN PROCESSED
        #####
        # ============================================================================>
        # Clear the linked maps if requested
        if not pandda.args.output.pickling.pickle_dataset_maps:
            for d_handler in pandda.datasets.mask(mask_name=analysis_mask_name):
                d_handler.child = None
        # Pickle the datasets
        pandda.sync_datasets(datasets=pandda.datasets.mask(mask_name=analysis_mask_name), overwrite_dataset_meta=True)
        pandda.pickle_the_pandda(components=['datasets'], datasets=pandda.datasets.mask(mask_name=analysis_mask_name))

        # ============================================================================>
        # DELETE THE MAP ANALYSER TO FREE UP MEMORY
        # ============================================================================>
        map_analyser.parent = None
        del map_analyser
        map_analyser = None
        # ============================================================================>
        # DELETE THE PROCESSED MAPS + Z-MAPS TO SAVE MEMORY (JUST BE CAREFUL NOT TO OVERWRITE THEIR PICKLED FILES!)
        # ============================================================================>
        for d_handler in pandda.datasets.mask(mask_name=analysis_mask_name):
            d_handler.child = None
        # ============================================================================>
        # Launch Garbage-Collection Manually just to be sure
        # ============================================================================>
        gc.collect()

        # ============================================================================>
        #####
        # LIVE ANALYSIS - RUN AT THE END OF EACH LOOP
        #####
        # Analyse the processed data
        # ============================================================================>
        # Extract all events from datasets
        all_events=[]; [all_events.extend(d.events) for d in pandda.datasets.all()]
        # Process any identified events and update output files
        if all_events:
            # ==================================================>
            # Cluster events and add to pandda tables
            # ==================================================>
            site_list = pandda.cluster_events_and_update(events=all_events)
            # ==================================================>
            # Update the output analysis files
            # ==================================================>
            pandda.write_output_csvs()
            analyse_html.write_analyse_html(pandda)
            # ==================================================>

    # ============================================================================>
    # Ensure that the collation step happens even if no resolutions are processed
    # ============================================================================>
    if not resolution_count:
        # Extract all events from datasets
        all_events=[]; [all_events.extend(d.events) for d in pandda.datasets.all()]
        # Process any identified events and update output files
        if all_events:
            # ==================================================>
            # Cluster events and add to pandda tables
            # ==================================================>
            site_list = pandda.cluster_events_and_update(events=all_events)
            # ==================================================>
            # Update the output analysis files
            # ==================================================>
            pandda.write_output_csvs()
            analyse_html.write_analyse_html(pandda)
            # ==================================================>

    # Pickle statistical maps again just to be sure
    pandda.pickle_the_pandda(components=['stat_maps'])

    # ============================================================================>
    #####
    # END OF MAIN LOOP
    #####
    # ============================================================================>
    t_analysis_end = time.time()
    # ==================================================>
    pandda.log('=====================================================>>>', True)
    pandda.log('Total Analysis Time: {!s}'.format(time.strftime("%H hours:%M minutes:%S seconds", time.gmtime(t_analysis_end - t_analysis_start))), True)
    pandda.log('=====================================================>>>', True)

    return

# ============================================================================>
#
###                 PanDDA Output + Wrap-Up Functions
#
# ============================================================================>

def pandda_end(pandda):
    """Write the output summaries and close"""

    pandda.log('',True)
    pandda.log('=====================================================>>>', True)
    pandda.log('Writing PanDDA End-of-Analysis Summary', True)
    pandda.log('=====================================================>>>', True)
    pandda.log('',True)

    # ============================================================================>
    #####
    # FINAL ANALYSIS - RUN ONCE AT THE END OF THE PROGRAM
    #####
    # Analyse all of the processed data
    # ============================================================================>

    # Collate events
    event_total, event_num, all_dataset_events = pandda.collate_event_counts()
    # Print some slightly less important information
    pandda.log('----------------------------------->>>', False)
    for d_tag, event_count in event_num:
        pandda.log('Dataset {!s}: {!s} Events'.format(d_tag, event_count), False)
    # Print a summary of the number of identified events
    pandda.log('----------------------------------->>>', True)
    pandda.log('Total Datasets with Events: {!s}'.format(len(event_num)), True)
    pandda.log('Total Events: {!s}'.format(event_total), True)

    pandda.log('----------------------------------->>>', True)
    pandda.log('Potentially Useful Shortcuts for Future Runs')
    if event_num:
        pandda.log('no_build={!s}'.format(','.join(zip(*event_num)[0])))

    # ============================================================================>
    #####
    # SUMMARIES ------------------------------>>>
    #####
    # ============================================================================>

    pandda.log('----------------------------------->>>', True)
    pandda.log('Writing final output files', True)
    pandda.write_output_csvs()
    analyse_html.write_analyse_html(pandda)

    # ============================================================================>
    # SCREEN GRAPHS -------------------------->>>
    # ============================================================================>

    resolution_counts = pandda.tables.dataset_map_info['analysed_resolution'].value_counts().sort_index()
    graph_data = [(str(r), c) for r,c in resolution_counts.iteritems()]
    if graph_data:
        try:
            from ascii_graph import Pyasciigraph
            g=Pyasciigraph()
            pandda.log('----------------------------------->>>', True)
            for l in g.graph(label='Datasets analysed at each resolution:', data=graph_data, sort=0):
                if l.startswith('#######'): continue
                pandda.log(l.replace(u"\u2588", '=').replace('= ','> '), True)
            pandda.log('----------------------------------->>>', True)
            pandda.log('')
        except ImportError: print('IMPORT ERROR (ascii_graph) - CANNOT GENERATE MAP ANALYSIS GRAPH')
        except:             pass
    else:
        pandda.log('> No Resolutions Analysed')

    pandda.log('=====================================================>>>', True)
    pandda.log('Datasets Processed: {!s}'.format(sum(resolution_counts)))
    pandda.log('Datasets Loaded {!s}'.format(pandda.datasets.size(mask_name='rejected - total', invert=True)))
    pandda.log('=====================================================>>>', True)

    return

# ============================================================================>
#
###                 PanDDA Command-Line Function
#
# ============================================================================>

def pandda_analyse_main(args):
    """Run the PANDDA algorithm, using supplied args"""

    working_phil = extract_params_default(master_phil=pandda_phil, args=args)

    try:
        # ============================================================================>
        #####
        # Initialise
        #####
        # ============================================================================
        pandda = PanddaMultiDatasetAnalyser(params=working_phil.extract())
        pandda.run_analysis_init()
        # ============================================================================>
        #####
        # Load and pre-process datasets. Generate grid and grid masks.
        #####
        # ============================================================================>
        pandda_dataset_setup(pandda=pandda)
        pandda_grid_setup(pandda=pandda)
        # ============================================================================>
        #####
        # Run the main analysis loop
        #####
        # ============================================================================>
        pandda_main_loop(pandda=pandda)
        # ============================================================================>
        #####
        # Write summaries and post-process
        #####
        # ============================================================================>
        pandda_end(pandda=pandda)
        # ============================================================================>
        #####
        # End
        #####
        # ============================================================================>
    except KeyboardInterrupt:
        raise
    except SystemExit:
        try:    pandda.log('Exited Normally')
        except: print '<<< Pandda exited before being initialised >>>'
    except:
        pandda.exit(error=True)
        raise
    else:
        pandda.exit(error=False)

    return pandda

# ============================================================================>
#
#   COMMAND LINE RUN
#
# ============================================================================>

if __name__ == '__main__':

    welcome()
    pandda = pandda_analyse_main(args=sys.argv[1:])






import os, time

from bamboo.wrappers.ligand_fitters import FitterObject

class FlynnObject(FitterObject):

    def _custom_init(self):
        self.program = 'flynn'
        self.name = 'flynn'

    def _custom_prepare_to_fit_ligand(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):
        """Prepare custom settings before fitting the ligand"""

        # Run the custom parts of this function
        self.outscores = self.outtemplate+'.scores'

    def _create_program_arguments(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):
        """Create command line and standard input to fit the ligand"""

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = ['-in',ligpdb,'-out',self.outtemplate+'.pdb','-prot',apopdb,'-map',mtz,'-prefix',self.outtemplate+'.auto','-reportfile',self.outscores]
        cmd_line_args.extend(cmd_line_flags)
        # Form Standard Input
        std_inpt_args = []
        std_inpt_args.extend(std_inpt_flags)

        return cmd_line_args, std_inpt_args

    def _post_process(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):

        # Get the output models
        self._find_output_files()

        return

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = ['-chainid', self.preferChain, '-split']
        std_inpt_flags = []

        return cmd_line_flags, std_inpt_flags

    def _find_output_files(self):
        """Reads the Flynn summary to find the output files"""

        workdir = self.outdir
        summary = self.outscores

        if not os.path.exists(summary):
            raise FittingError('No Fitting Summary File Found: {!s}\n\t\tLog: {!s}'.format(summary, self.outlog))

        summary_contents = open(summary,'r').readlines()

        fields = []
        rows = []
        data = {}

        for i, line in enumerate(summary_contents):
            line=line.strip()
            if line.startswith('Smiles'):
                fields = line.split(',')
            elif fields:
                rows.append(line.split(','))

        if not (fields and rows):
            raise FittingError('No Fitted Files Found in Summary File: {!s}\n\t\tLog: {!s}'.format(summary, self.outlog))

        for i, row in enumerate(rows):
            # Create dictionary of score information
            rowdict = dict(zip(fields,row))
            # Create a rank (the files are already ordered by 'rank' from flynn)
            rank = i+1
            # Create filename
            filename = '{!s}_n{:0>3}_b{:0>3}_s{:0>2}_c{!s:0>3}.pdb'.format(self.outtemplate,str(rank),rowdict['Blobs'],rowdict['Stereo Variant'],int(rowdict['Conformer'])-1)
            # Add a couple more fields
            rowdict['Rank'] = rank
            rowdict['File'] = filename
            rowdict['Path'] = os.path.join(self.outdir,filename)
            # Create a common output score
            rowdict['InternalScore'] = rowdict['RSCC']
            # Check path exists
            assert os.path.exists(rowdict['Path']), 'Flynn Output File does not exist: {!s}'.format(rowdict)
            # Add to list of output model data (ordered by rank from flynn)
            data[rowdict['Path']] = rowdict
            # Add path to list of output models (ordered by rank from flynn)
            self.outmodels.append(rowdict['Path'])

        # Store scoring data
        self.fittingdata = data

        return





from bamboo.stats.cluster import convert_tree_to_edges

def plot_tree_as_network(cluster_node):
    """Takes a scipy ClusterNode object and plot as a phylogenetic tree"""

    edges = convert_tree_to_edges(cluster_node)





import os, sys, copy, re, glob

import libtbx.phil
import libtbx.easy_mp

import numpy, pandas

from bamboo.html import BAMBOO_HTML_ENV
from giant.jiffies.score_model import prepare_output_directory, score_model, make_residue_radar_plot, format_parameters_for_plot
from giant.stats.cluster import generate_group_idxs

from matplotlib import pyplot
pyplot.rc('axes', color_cycle=['r', 'g', 'b', 'y'])

#######################################

bar = '=======================++>'

#######################################

blank_arg_prepend = {None:'pdb='}

master_phil = libtbx.phil.parse("""
input {
    pdb = None
        .type = path
        .multiple = True
    ref_pdb = None
        .type = path
        .multiple = False
    labels = basename *folder_name
        .type = choice
        .multiple = False
}
selection {
    res_names = LIG,UNL,DRG
        .type = str
        .help = "Comma-separated list of residue names to score -- if None then scores all"
}
output {
    out_dir = ./
        .type = path
    generate_summary = True
        .type = bool
}
radar_plot {
    limits = automatic *manual
        .help = 'Select how the axis limits are determined'
        .type = choice
        .multiple = False
}
include scope giant.jiffies.score_model.residue_plot_phil
""", process_includes=True)

#######################################

def run(params):

    assert params.input.pdb, 'No pdb files provided'
    # REMOVE THIS WHEN IMPLEMENTED
    assert params.selection.res_names
    # REMOVE THIS WHEN IMPLEMENTED
    if params.selection.res_names:      params.selection.__inject__("res_names_list", params.selection.res_names.split(','))
    else:                               params.selection.__inject__("res_names_list", None)

    output_dir, images_dir = prepare_output_directory(params)
    scores_file =  os.path.join(output_dir, 'residue_scores.csv')

    all_data = pandas.DataFrame()

    for pdb in params.input.pdb:

        mtz   = pdb.replace('.pdb','.mtz')

        if   params.input.labels == 'basename':    label = os.path.split_ext(os.path.basename(pdb))[0]
        elif params.input.labels == 'folder_name': label = os.path.basename(os.path.dirname(os.path.abspath(pdb)))

        print bar
        print 'Scoring model {} against {}'.format(pdb, mtz)

        data_table = score_model(   params = params,
                                    pdb1   = pdb,
                                    mtz1   = mtz,
                                    pdb2   = params.input.ref_pdb,
                                    label_prefix = label
                                )
        all_data = all_data.append(data_table, verify_integrity=True)

    print '...Done'
    print bar

    all_data.to_csv(scores_file)
    print 'Output written to {}'.format(scores_file)
    print bar

    ###################################################################
    # Image parameters
    ###################################################################
    columns = format_parameters_for_plot(params=params.plot.parameters)

    ###################################################################
    # Output Images - 1 image per residue per structure
    ###################################################################
    all_images = []
    print 'Generating Output Images...'
    for label, row in all_data.iterrows():
        image_path = os.path.join(images_dir,'{}.png'.format(label))
        print 'Making: {}...'.format(image_path)
        make_residue_radar_plot(path = image_path,
                                data = row.to_frame().T,
                                columns = columns,
                                remove_blank_entries = params.plot.remove_blank_entries,
                                print_axis_values    = params.plot.print_axis_values  )
        all_images.append(image_path)

    ###################################################################
    # Output Images - 1 image per residue (allowing comparisons)
    ###################################################################
    if params.radar_plot.limits == 'automatic': columns.pop('limits', None)
    elif params.radar_plot.limits == 'manual':  pass

    for res_label, index_idxs in generate_group_idxs([i.split('-')[-3:] for i in all_data.index]):
        res_label = '-'.join(res_label)
        image_path = os.path.join(images_dir,'compare-{}.png'.format(res_label))
        print 'Making: {}...'.format(image_path)
        make_residue_radar_plot(path = image_path,
                                data = all_data.irow(index_idxs),
                                columns = columns,
                                remove_blank_entries = params.plot.remove_blank_entries,
                                print_axis_values    = params.plot.print_axis_values  )

    print '...Done.'
    print bar

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, sys, shutil

from bamboo.common.command import CommandManager
from bamboo.rdkit_utils.mol import check_pdb_readable

def generate_symmetry_mates(pdbin, pdbout, sgno, cell):
    """Takes the input pdb and generates the unit cell from the point group symmetry"""

    if not pdbout.endswith('.pdb'):
        pdbout = pdbout+'.pdb'
    if not os.path.exists(pdbin):
        raise IOError('pdbin does not exist! {!s}'.format(pdbin))
    if os.path.exists(pdbout):
        raise Exception('pdbout already exists! {!s}'.format(pdbout))

    assert isinstance(sgno, int), 'SPACEGROUP MUST BE AN INTEGER! {!s}'.format(sgno)
    assert isinstance(cell, list), 'CELL MUST BE A LIST! {!s}'.format(cell)

    # Initialise Commander
    PDBSET = CommandManager('pdbset')
    # Set Command Arguments
    PDBSET.add_command_line_arguments('XYZIN',os.path.abspath(pdbin),'XYZOUT',os.path.abspath(pdbout))
    # Set inputs
    PDBSET.add_standard_input(['SYMGEN {!s}'.format(sgno),'CELL {!s}'.format(' '.join(map(str,cell))), 'END'])
    # run!
    PDBSET.run()

    if not os.path.exists(pdbout):
        raise ExternalProgramError('PDBSET has failed to generate SYMMETRY mates. {!s}\nCOM: {!s}\nOUT: {!s}\nERR: {!s}'.format(pdbin,PDBSET.command,PDBSET.out,PDBSET.err))

    return pdbout

def map_to_reference_using_symmetry(refpdb, movpdb, pdbout, conrad=5):
    """Transforms `movpdb` to the closest symmetry site to `refpdb` using csymmatch - Symmetry info must be contained in the header of the pdb file"""

    if not pdbout.endswith('.pdb'):
        pdbout = pdbout+'.pdb'
    if not os.path.exists(refpdb):
        raise IOError('refpdb does not exist! {!s}'.format(refpdb))
    if not os.path.exists(movpdb):
        raise IOError('movpdb does not exist! {!s}'.format(movpdb))
    if os.path.exists(pdbout):
        raise Exception('pdbout already exists! {!s}'.format(pdbout))

    # Initialise Commander
    CSYMMATCH = CommandManager('csymmatch')
    # Set Command Arguments
    CSYMMATCH.add_command_line_arguments('-pdbin-ref',os.path.abspath(refpdb),'-pdbin',os.path.abspath(movpdb),'-pdbout',os.path.abspath(pdbout),'-connectivity-radius',str(conrad))
    # run!
    CSYMMATCH.run()

    if not os.path.exists(pdbout):
        raise ExternalProgramError('CSYMMATCH has failed to map {!s} to {!s}.\nERR: {!s}'.format(movpdb,refpdb,CSYMMATCH.err))

    # Check output is valid (raises LigandCheckError if it fails)
    check_pdb_readable(pdbout)

    return pdbout



from rdkit import Chem

from bamboo.common.file import FileObj
from bamboo.macro.molecule import MacroMol
from bamboo.rdkit_utils.mol import check_pdb_readable
from bamboo.constants import ATOM_NUMBER_DICT

################################################################################################################
### CLASSES                                                                                                    #
################################################################################################################

class LigFile(object):
    """Class for summarising a ligand PDBFile"""
    def __init__(self, ligpath, parse=False):
        self.file = FileObj(ligpath)
        # Convert to mol, checking that's it's valid chemically
        self.mol = check_pdb_readable(ligpath)
        # Create MacroMol object
        if parse:
            self.macromol = MacroMol(ligpath)
        else:
            self.macromol = None

    def get_total_atom_count(self):
        return self.mol.GetNumAtoms()

    def get_heavy_atom_count(self):
        return self.mol.GetNumHeavyAtoms()

    def get_light_atom_count(self):
        return self.mol.GetNumAtoms() - self.mol.GetNumHeavyAtoms()

    def get_bond_count(self):
        return self.mol.GetNumBonds()

    def get_conformer_count(self):
        return self.mol.GetNumConformers()

    def get_atom_counts(self):
        """Gets a summary of the counts of each atom in the ligand"""

        atomcounts = {}
        ats = self.mol.GetAtoms()
        for a in ats:
            atomic_symbol = ATOM_NUMBER_DICT[a.GetAtomicNum()]
            if atomic_symbol not in atomcounts:
                atomcounts[atomic_symbol] = 1
            else:
                atomcounts[atomic_symbol] += 1

        return atomcounts

    def get_atom_counts_by_fragment(self):
        """Gets atom counts by residue"""

        fragments = {}

        inscodes = sorted(list(set([at.GetMonomerInfo().GetInsertionCode() for at in self.mol.GetAtoms()])))

        for ins in inscodes:
            ats = [at for at in self.mol.GetAtoms() if at.GetMonomerInfo().GetInsertionCode()==ins]

            atomcounts = {}
            for a in ats:
                atomic_symbol = ATOM_NUMBER_DICT[a.GetAtomicNum()]
                if atomic_symbol not in atomcounts:
                    atomcounts[atomic_symbol] = 1
                else:
                    atomcounts[atomic_symbol] += 1

            fragments[ins] = atomcounts

        return fragments

    def get_total_atom_count_by_fragment(self):

        fragments = {}

        inscodes = sorted(list(set([at.GetMonomerInfo().GetInsertionCode() for at in self.mol.GetAtoms()])))

        for ins in inscodes:
            ats = [at for at in self.mol.GetAtoms() if at.GetMonomerInfo().GetInsertionCode()==ins]
            fragments[ins] = len(ats)

        return fragments




import os, sys, copy

import libtbx.phil

import numpy

import iotbx.pdb

from scitbx.array_family import flex

from giant.structure.occupancy import normalise_occupancies, set_conformer_occupancy

#######################################

blank_arg_prepend = None

master_phil = libtbx.phil.parse("""

conf_ids = None
    .type = str
res_ids = LIG,UNL,FRG
    .type = str

set_occ = 0.1
    .type = float

input = None
    .type = path
output = None
    .type = path

minimum_occ = 0.0
    .type = float
maximum_occ = 1.0
    .type = float

overwrite = False
    .type = bool
verbose = False
    .type = bool
""")

#######################################

def run(params):

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'SETTING OCCUPANCIES'
    ######################################################################

    assert params.input, 'No Input file given!'

    if not params.output: params.output='./occ-edited.pdb'
    if os.path.exists(params.output):
        if params.overwrite: os.remove(params.output)
        else: raise Exception('File already exists: {}'.format(params.output))

    assert params.conf_ids or params.res_ids, 'Must provide conformer IDs or residue IDs'

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'READING INPUT FILES'
    ######################################################################
    # Read in the ligand file and set each residue to the requested conformer
    prot_h = iotbx.pdb.hierarchy.input(params.input).hierarchy

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'VALIDATING INPUT MODELS'
    ######################################################################
    # Check that ... something
    prot_h.only_model()

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'SETTING OCCUPANCIES'
    ######################################################################

    if params.res_ids:  sel_res_ids = params.res_ids.split(',')
    else:               sel_res_ids = []
    if params.conf_ids: sel_conf_ids = params.conf_ids.split(',')
    else:               sel_conf_ids = []
    all_conf_ids = sel_conf_ids + list(set([ag.altloc for ag in prot_h.atom_groups() if ag.altloc and (ag.resname in sel_res_ids)]))

    print 'Setting Occupancy for Conformers: {}'.format(sorted(all_conf_ids))

    for conf in all_conf_ids:
        if params.verbose: print 'Setting Conformer {} to {}'.format(conf, params.set_occ)
        set_conformer_occupancy(hierarchy=prot_h, conf_id=conf, conf_occ=params.set_occ, in_place=True)

    prot_h = normalise_occupancies( hierarchy = prot_h,
                                    exclude_conformers = all_conf_ids,
                                    max_occ = params.maximum_occ,
                                    min_occ = params.minimum_occ  )

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'WRITING OUTPUT STRUCTURE'
    ######################################################################

    # Write output file
    prot_h.write_pdb_file(params.output)

    return

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os,sys,re

from bamboo.constants import MAP_TYPES

# Contains Functions for Manipulating MTZs

def select_factors_and_phases_for_map(cols, maptype='2FOFC'):
    """Takes the Column Headings and returns a dictionary for creating a map type"""

    # Check for Valid Map
    if maptype not in MAP_TYPES:
        raise Exception('MapType not in allowed map types: {!s} not in {!s}'.format(maptype,MAP_TYPES))
    # Dictionary to contain headings
    dict = {'F':'','PHI':''}
    # Match
    if maptype == '2FOFC':
        if ('2FOFCWT' in cols) and ('PH2FOFCWT' in cols):
            dict['F'] = '2FOFCWT'
            dict['P'] = 'PH2FOFCWT'
        elif ('FWT' in cols) and ('PHWT' in cols):
            dict['F'] = 'FWT'
            dict['P'] = 'PHWT'
        elif ('FWT' in cols) and ('PHFWT' in cols):
            dict['F'] = 'FWT'
            dict['P'] = 'PHFWT'
        else:
            raise Exception('Failed to select STRUCTURE FACTOR Columns for Map Type {!s}. \nColumns : {!s}'.format(maptype,cols))
    elif maptype == 'FOFC':
        if ('FOFCWT' in cols) and ('PHFOFCWT' in cols):
            dict['F'] = 'FOFCWT'
            dict['P'] = 'PHFOFCWT'
        elif ('DELFWT' in cols) and ('DELPHWT' in cols):
            dict['F'] = 'DELFWT'
            dict['P'] = 'DELPHWT'
        elif ('DELFWT' in cols) and ('PHDELWT' in cols):
            dict['F'] = 'DELFWT'
            dict['P'] = 'PHDELWT'
        else:
            raise Exception('Failed to select PHASE Columns: {!s}'.format(cols))
    else:
        raise Exception('MapType does not have functionality programmed: {!s}'.format(maptype))
    # Return
    return dict




import os, sys, copy, re, glob

import libtbx.phil
import libtbx.easy_mp

import numpy, pandas

from bamboo.html import BAMBOO_HTML_ENV
from giant.jiffies.score_model import prepare_output_directory, score_model, make_residue_radar_plot, format_parameters_for_plot, residue_plot_phil

#######################################

bar = '=======================++>'

#######################################

blank_arg_prepend = {None:'dir='}

master_phil = libtbx.phil.parse("""
input {
    dir = None
        .type = path
        .multiple = True

    pdb_1_style = 'refine.pdb'
        .type = str
    mtz_1_style = 'refine.mtz'
        .type = str

    pdb_2_style = '*-ensemble-model.pdb'
        .type = str
    mtz_2_style = '*-input.mtz'
        .type = str

    labels = basename *folder_name
        .type = choice
        .multiple = False
}
selection {
    res_names = LIG,UNL,DRG
        .type = str
        .help = "Comma-separated list of residue names to score -- if None then scores all"
}
output {
    out_dir = ./
        .type = path
    generate_summary = True
        .type = bool
}
settings{
    cpus = 1
        .type = int
}
include scope giant.jiffies.score_model.residue_plot_phil
""", process_includes=True)

#######################################

def process_folder_mp(args):
    try:
        return process_folder(*args)
    except Exception as e:
        return e

def process_folder(dir, params):
    """Score the files in a folder"""

    try:    pdb1 = glob.glob(os.path.join(dir, params.input.pdb_1_style))[0]
    except: raise Exception('FAILED TO FIND FILE: {} in {}'.format(params.input.pdb_1_style, dir))
    try:    mtz1 = glob.glob(os.path.join(dir, params.input.mtz_1_style))[0]
    except: raise Exception('FAILED TO FIND FILE: {} in {}'.format(params.input.mtz_1_style, dir))
    if params.input.pdb_2_style:
        try:    pdb2 = glob.glob(os.path.join(dir, params.input.pdb_2_style))[0]
        except: raise Exception('FAILED TO FIND FILE: {} in {}'.format(params.input.pdb_2_style, dir))
    else: pdb2 = None
    if params.input.mtz_2_style:
        try:    mtz2 = glob.glob(os.path.join(dir, params.input.mtz_2_style))[0]
        except: raise Exception('FAILED TO FIND FILE: {} in {}'.format(params.input.mtz_2_style, dir))
    else: mtz2 = None

    if   params.input.labels == 'basename':     label = os.path.split_ext(os.path.basename(pdb1))[0]
    elif params.input.labels == 'folder_name':  label = os.path.basename(dir)

    print 'Scoring Model: {}'.format(label)

    data_table = score_model(   params = params,
                                pdb1   = pdb1,
                                mtz1   = mtz1,
                                pdb2   = pdb2,
                                mtz2   = mtz2,
                                label_prefix = label
                            )

    return data_table

#######################################

def run(params):

    assert params.input.dir, 'No directories provided'
    assert params.input.pdb_1_style
    assert params.input.mtz_1_style
    # REMOVE THIS WHEN IMPLEMENTED
    assert params.selection.res_names
    # REMOVE THIS WHEN IMPLEMENTED
    if params.selection.res_names:      params.selection.__inject__("res_names_list", params.selection.res_names.split(','))
    else:                               params.selection.__inject__("res_names_list", None)

    output_dir, images_dir = prepare_output_directory(params)
    scores_file  = os.path.join(output_dir, 'residue_scores.csv')
    summary_file = os.path.join(output_dir, 'residue_scores.html')

    arg_list = []
    # Filter directories
    for dir in params.input.dir:
        if not os.path.isdir(dir): continue
        # Build arglist for mapping
        arg_list.append((dir, params))
    # Score files in parallel
    print bar
    print 'Scoring all models...'
    returned_objs = libtbx.easy_mp.pool_map(fixed_func=process_folder_mp, args=arg_list, processes=params.settings.cpus)
    # Extract results
    data_table = None
    for obj in returned_objs:
        if isinstance(obj, pandas.DataFrame):
            if data_table is None:  data_table = obj
            else:                   data_table = data_table.append(obj, verify_integrity=True)
        else:
            try:    print '{}: {}'.format(obj.__class__, obj.message)
            except: print 'Unknown Error: {}'.format(str(obj))
    print '...Done'
    print bar

    data_table.to_csv(scores_file)
    print 'Output written to {}'.format(scores_file)
    print bar

    ###################################################################
    # Image parameters
    ###################################################################
    columns = format_parameters_for_plot(params=params.plot.parameters)

    ###################################################################
    # Output Images
    ###################################################################
    all_images = []
    print 'Generating Output Images...'
    for label, row in data_table.iterrows():
        image_path = os.path.join(images_dir,'{}.png'.format(label))
        print 'Making: {}...'.format(image_path)
        make_residue_radar_plot(path = image_path,
                                data = row.to_frame().T,
                                columns = columns,
                                remove_blank_entries = params.plot.remove_blank_entries,
                                print_axis_values    = params.plot.print_axis_values  )
        all_images.append(image_path)
    print '...Done'
    print bar

    if params.output.generate_summary:
        print 'Generating output HTML...'
        # Get template to be filled in
        template = BAMBOO_HTML_ENV.get_template('summary_page.html')
        # Output directory (for relative symlinks)
        out_dir  = os.path.abspath(os.path.dirname(summary_file))

        # ===========================================================>
        # Construct the data object to populate the template
        output_data = {}
        output_data['header'] = 'Residue Score Summaries'
        output_data['title'] = 'Residue Score Summaries'
        output_data['introduction'] = 'Model Quality and Validation checks.'
        # ===========================================================>
        # Header Images
        output_data['small_images'] = []
        for img in all_images:
            output_data['small_images'].append({ 'path': './'+os.path.relpath(path=img, start=out_dir),
                                                 'title': 'Scores for {}'.format(os.path.splitext(os.path.basename(img))[0]) })
        # ===========================================================>
        # Write Output
        with open(summary_file, 'w') as out_html:
            out_html.write(template.render(output_data))
        print '...Done'
        print bar

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, sys, copy

import libtbx.phil

import numpy

import iotbx.pdb

from scitbx.array_family import flex

from giant.utils.pdb import strip_pdb_to_input, get_pdb_header

#######################################

blank_arg_prepend = {'.pdb':'pdb='}

master_phil = libtbx.phil.parse("""
pdb = None
    .help = 'The ensemble of protein conformations (normally the unbound and bound structures)'
    .type = str
    .multiple = True

res = LIG,UNL
    .help = 'Residues to be define selected conformations (comma separated list of residue names, i.e. res=LIG or res=LIG,UNL)'
    .type = str
conf = None
    .help = 'Define selected conformations explicitly (comma separated list of conformer IDs, i.e. conf=A or conf=A,B)'
    .type = str

suffix = '.stripped.pdb'
    .help = 'output suffix for input files'
    .type = str

reset_altlocs = True
    .help = 'Relabel conformers of kept residues'
    .type = bool

overwrite = False
    .type = bool
verbose = False
    .type = bool

log = None
    .type = path
    .multiple = False
""")

#######################################

def proc(ensemble_file, params, sel_resnames=None, sel_confs=None):

    # Create output file path
    output_file = os.path.splitext(ensemble_file)[0] + params.suffix

    # Read the pdb header
    header_contents = get_pdb_header(ensemble_file)

    # Read in the ligand file and set each residue to the requested conformer
    ens_obj = strip_pdb_to_input(ensemble_file, remove_ter=True)

    # Check that ... something
    ens_obj.hierarchy.only_model()

    # Create a new copy of the structures
    new_ens = ens_obj.hierarchy.deep_copy()

    confs_to_select = set()
    if params.res:
        confs_to_select = confs_to_select.union([ag.altloc for ag in new_ens.atom_groups() if (ag.resname in sel_resnames) and ag.altloc])
    if params.conf:
        confs_to_select = confs_to_select.union(sel_confs)
    confs_to_select = sorted(confs_to_select)

    print 'Keeping Conformer IDs: {}'.format(', '.join(confs_to_select))

    if not confs_to_select: raise Exception('No Conformers Selected')

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'PRUNING OUTPUT STRUCTURE'
    ######################################################################

    print '===========================================>>>'
    print 'Keeping ANY atom with conformer id: {}'.format(' or '.join(['" "']+confs_to_select))
    print 'Keeping ALL amino acid conformers with no conformer in {}'.format(' or '.join(confs_to_select))
    print 'Removing amino acid conformers where conformer not in {} but there is a conformer in this'.format(' or '.join(confs_to_select))
    print 'Removing any other atom where conformer id not in {}'.format(' or '.join(confs_to_select))
    print '===========================================>>>'

    # Select atoms to keep - no altloc, or altloc in selection
    sel_string_1 = ' or '.join(['altid " "']+['altid {}'.format(alt) for alt in confs_to_select])
    print 'Selection 1: \n\t'+sel_string_1

    # Select amino acids with no conformer overlapping with confs_to_select
    sel_ags = [ag for ag in new_ens.atom_groups() if (iotbx.pdb.common_residue_names_get_class(ag.resname)=='common_amino_acid') and not set(confs_to_select).intersection([a.altloc for a in ag.parent().atom_groups()])]
    sel_string_2 = ' or '.join(['(chain {!s} and resid {!s} and altid "{!s}")'.format(ag.parent().parent().id, ag.parent().resid(), ag.altloc) for ag in sel_ags])
    print 'Selection 2: \n\t'+sel_string_2.replace('or ','or\n\t')[:500]+'...'

    sel_cache = new_ens.atom_selection_cache()
    sel_string = ' or '.join(['({})'.format(s) for s in [sel_string_1, sel_string_2]])

    conf_sel = new_ens.select(sel_cache.selection(sel_string))

    if params.log:
        fh = open(params.log, 'w')
        fh.write('>>> Residues where the conformer will be reset\n')
    else:
        fh = None

    if params.reset_altlocs:
        ######################################################################
        if params.verbose: print '===========================================>>>'
        if params.verbose: print 'Resetting OUTPUT STRUCTURE'
        ######################################################################

        if len(confs_to_select)==1: hash = {confs_to_select[0]: ' '}
        else:                       hash = dict(zip(confs_to_select, iotbx.pdb.systematic_chain_ids()))

        if params.verbose:
            for it_pr in hash.items(): print 'Changing ALTLOCs:', ' -> '.join(it_pr)

        for ag in conf_sel.atom_groups():
            if ag.altloc in confs_to_select:
                # Reset the altloc
                ag.altloc = hash[ag.altloc]
                # Log this atom group
                if fh: fh.write('> Chain {}, Residue {}, (New) Conformer {}\n'.format(ag.parent().parent().id, ag.parent().resid(), ag.altloc))

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'WRITING OUTPUT STRUCTURE'
    ######################################################################

    # Update the atoms numbering
#        conf_sel.atoms_reset_serial()
    with open(output_file, 'w') as fh:
        fh.write(header_contents)
    # Write output file
    conf_sel.write_pdb_file(output_file, open_append=True)

#######################################

def run(params):

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'VALIDATING INPUT PARAMETERS'
    ######################################################################

    assert params.pdb, 'No PDB files given'
    assert params.res or params.conf, 'Must give either res or conf'

    ######################################################################
    if params.verbose: print '===========================================>>>'
    if params.verbose: print 'SPLITTING ENSEMBLES'
    ######################################################################

    if params.res:
        sel_resnames = params.res.split(',')
    else:
        sel_resnames = None

    if params.conf:
        sel_confs = params.conf.split(',')
    else:
        sel_confs = None

    # Iterate through the input structures and extract the conformation
    for ensemble_file in params.pdb:
        try:
            proc(ensemble_file, params, sel_resnames=sel_resnames, sel_confs=sel_confs)
        except:
            print "Failed:", ensemble_file
            raise

    return

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import numpy

from libtbx.math_utils import iceil, ifloor

def interpolate(x,p1,v1,p2,v2):
    """Interpolate the values of p1 and p2 to the point at x"""
    if x==p1: return v1
    if x==p2: return v2
    return v2*(x-p1)/(p2-p1) + v1*(p2-x)/(p2-p1)

def calculate_minimum_redundancy(unsampled_size, max_sample_size):
    """Calculate the redundancy required to down-sample a vector of size unsampled_size to size max_sample_size"""
    min_redundancy = int(1+(unsampled_size-1)/max_sample_size)
    return min_redundancy

def resample_ordered_list_of_values(vals, redundancy=8):
    """resample a list of values with interpolation"""
    # Number of vals given
    num_inp_vals = len(vals)
    # Number of vals to be returned
    num_samp_vals = int(1+(num_inp_vals-1)/redundancy)
    # Sort in descending order
    ordered_vals = sorted(vals, reverse=True)
    sampled_vals = []

    if num_samp_vals==1:
        return [ordered_vals[0]]
    else:
        sample_dist = (num_inp_vals-1)/(num_samp_vals-1)
#        # Make sure it doesn't overrun the end of the array
#        while sample_dist*(num_samp_vals-1) > num_inp_vals-1:
#            sample_dist = 0.99999999*sample_dist

    # Resample points with interpolation
    for i_point in range(num_samp_vals-1):
        sample_index = sample_dist*i_point
        p1 = ifloor(sample_index)
        v1 = ordered_vals[p1]
        p2 = iceil(sample_index)
        v2 = ordered_vals[p2]
        sample_val = interpolate(x=sample_index, p1=p1, v1=v1, p2=p2, v2=v2)
        sampled_vals.append(sample_val)
    # Add the last point
    sampled_vals.append(ordered_vals[-1])

    assert len(sampled_vals) == num_samp_vals

    return sampled_vals

def calculate_roc_vals(correct_class, rank_vals):
    """Calculate the ROC Curve values for the ordered outcomes"""

    assert len(correct_class) == len(rank_vals)

    # Sort rank vals into descending order
    sorted_indices = sorted(range(len(rank_vals)), key=lambda k: rank_vals[k], reverse=True)
    # Use this to reorder
    sorted_correct_class = [correct_class[i] for i in sorted_indices]
    sorted_rank_vals =     [rank_vals[i] for i in sorted_indices]

    # Calculate the list of cutoffs
    cutoffs = numpy.zeros(len(rank_vals) + 1)
    cutoffs[0] =  max(sorted_rank_vals) + 1
    cutoffs[-1] = min(sorted_rank_vals) - 1
    cutoffs[1:-1] = (numpy.array(sorted_rank_vals[1:]) + numpy.array(sorted_rank_vals[:-1])) / 2.
    cutoffs = cutoffs.tolist()

    T_BOOL = [1.0 if t     else 0.0 for t in sorted_correct_class]
    N_BOOL = [1.0 if not t else 0.0 for t in sorted_correct_class]
    assert sum(T_BOOL) + sum(N_BOOL) == len(sorted_correct_class)

    # Iterate through the data and calculate TP, TN, FP, FN Rates
    POS_TOTAL = sum(T_BOOL)
    NEG_TOTAL = sum(N_BOOL)
    TOTAL = POS_TOTAL + NEG_TOTAL

    # INITIAL VALUES - STARTING FROM CLASSIFYING EVERYTHING AS NEGATIVE: NO POSITIVES, ALL NEGATIVES
    TPR = [0.0]; FPR = [0.0]; TNR = [1.0]; FNR = [1.0];
    PPV = [0.0]; NPV = [NEG_TOTAL/TOTAL];
    ACC = [NEG_TOTAL/TOTAL]
    TP  = [0.0]; FP  = [0.0]; TN  = [NEG_TOTAL]; FN  = [POS_TOTAL];

    for index in range(len(correct_class)):
        tp = sum(T_BOOL[0:index+1])
        fp = sum(N_BOOL[0:index+1])
        tn = sum(N_BOOL[index+1:] )
        fn = sum(T_BOOL[index+1:] )
        assert tp + fn == POS_TOTAL, 'POS TOTAL != TP + FN'
        assert tn + fp == NEG_TOTAL, 'NEG TOTAL != TN + FP'

#        print 'TP', tp, 'TN', tn, 'FP', fp, 'FN', fn

        TP.append(tp)
        FP.append(fp)
        TN.append(tn)
        FN.append(fn)

        # Fraction of positives classed as hits
        TPR.append(tp/(tp+fn))
        # Fraction of negatives classed as hits
        FPR.append(fp/(fp+tn))

        # Fraction of negatives classed as negs
        TNR.append(tn/(tn+fp))
        # Fraction of positives classed as negs
        FNR.append(fn/(fn+tp))

        # Fraction of hits that are positive
        try:                        PPV.append(tp/(tp+fp))
        except ZeroDivisionError:   PPV.append(1.0)
        # Fraction of negs that are negative
        try:                        NPV.append(tn/(tn+fn))
        except ZeroDivisionError:   NPV.append(1.0)

        # Fraction of classifications that are correct
        ACC.append((tp+tn)/TOTAL)

    SENS = TPR
    SPEC = TNR
    PREC = PPV

    ROC_RESULTS =   {
                    'VALS'  : cutoffs,
                    'TP'    : TP,       'FP'    : FP,
                    'TN'    : TN,       'FN'    : FN,
                    'TPR'   : TPR,      'FPR'   : FPR,
                    'TNR'   : TNR,      'FNR'   : FNR,
                    'PPV'   : PPV,      'NPV'   : NPV,
                    'SENS'  : SENS,     'SPEC'  : SPEC,
                    'PREC'  : PREC,     'ACC'   : ACC
                    }

    return ROC_RESULTS



import os, sys
import base64, hashlib
import socket
import threading

# CONNECTION CLASSES

class ConnectionReferenceObject(object):
    """ Object for storing socket information """
    def __init__(self):
        self.sock = None
        self.conn = None
        self.lock = None
        self.addr = None

class SocketReferenceObject(object):
    """ Object for passing Incoming and Outgoing Socket Objects """
    def __init__(self):
        """Spaces for incoming/outgoing sockets, connections, locks"""
        self.incoming = ConnectionReferenceObject()
        self.outgoing = ConnectionReferenceObject()

class ThreadWithReturnValue(threading.Thread):
    """ Define Thread so that it has a return value """
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, verbose=False):
        threading.Thread.__init__(self, group, target, name, args, kwargs, verbose)
        self._return = None
    def run(self):
        if self._Thread__target is not None:
            self._return = self._Thread__target(*self._Thread__args,
                                                **self._Thread__kwargs)
    def join(self):
        threading.Thread.join(self)
        return self._return

# CONNECTION FUNCTIONS

def create_handshake_resp(handshake):
    """ Respond to handshake to establish connection """
    final_line = ""
    lines = handshake.splitlines()
    for line in lines:
        parts = line.partition(": ")
        if parts[0] == "Sec-WebSocket-Key":
            key = parts[2]

    magic = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
    accept_key = base64.b64encode(hashlib.sha1(key+magic).digest())

    return (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: WebSocket\r\n"
        "Connection: Upgrade\r\n"
        "Sec-WebSocket-Accept: " + accept_key + "\r\n\r\n")

# DECODING FUNCTIONS

def decode_message(string_stream_in):
    """ Decode a WebSocket data frame into a string """

    #turn string values into opererable numeric byte values
    byteArray = [ord(character) for character in string_stream_in]
    datalength = byteArray[1] & 127
    indexFirstMask = 2
    if datalength == 126:
        indexFirstMask = 4
    elif datalength == 127:
        indexFirstMask = 10
    masks = [m for m in byteArray[indexFirstMask : indexFirstMask+4]]
    indexFirstDataByte = indexFirstMask + 4
    decoded_chars = []
    i = indexFirstDataByte
    j = 0
    while i < len(byteArray):
        decoded_chars.append( chr(byteArray[i] ^ masks[j % 4]) )
        i += 1
        j += 1

    return ''.join(decoded_chars)

def encode_message_to_send(raw_data):
    """Turns a string into a WebSocket data frame. Returns a bytes(). 'data' is a string"""

    #determine the size of the data we were told to send
    data_length = len(raw_data)
    output_bytes = bytearray()
    output_bytes.append(0x81) #0x81 = text data type
    if data_length < 0x7D:
        #a nice short length
        output_bytes.append(len(raw_data))
    elif data_length >= 0x7E and len(raw_data) < 0xFFFF:
        #two additional bytes of length needed
        output_bytes.append(0x7E)
        output_bytes.append(data_length >> 8 & 0xFF)
        output_bytes.append(data_length & 0xFF)
    else:
        #eight additional bytes of length needed
        output_bytes.append(0x7F)
        output_bytes.append(data_length >> 56 & 0xFF)
        output_bytes.append(data_length >> 48 & 0xFF)
        output_bytes.append(data_length >> 40 & 0xFF)
        output_bytes.append(data_length >> 32 & 0xFF)
        output_bytes.append(data_length >> 24 & 0xFF)
        output_bytes.append(data_length >> 16 & 0xFF)
        output_bytes.append(data_length >> 8 & 0xFF)
        output_bytes.append(data_length & 0xFF)
    #tack on the raw data now
    for byte in raw_data:
        output_bytes.append(ord(byte))
    return bytes(output_bytes)

def send_message_to_connection(message, connection):
    """ Message is a String. Connection is a ConnectionReferenceObject. """
    msg = encode_message_to_send(message)
    connection.lock.acquire()
    connection.conn.send(msg)
    connection.lock.release()

def establish_outgoing_connection(outgoing_connection, ADDR, PORT):
    """ Establish a connection with COOT """

    print "Outgoing Connection: ", ADDR, ' on ', PORT
    print "WAITING FOR OUTGOING CONNECTION TO BE MADE"

    outgoing_connection.sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    outgoing_connection.sock.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
    outgoing_connection.sock.bind((ADDR,PORT))
    outgoing_connection.sock.listen(1)

    # Handle Request
    outgoing_connection.conn, outgoing_connection.addr = outgoing_connection.sock.accept()
    print 'Outgoing Connection made to: ', outgoing_connection.addr
    data = outgoing_connection.conn.recv(1024)
    print 'Received Msg: ', data

    return outgoing_connection

def establish_incoming_connection(incoming_connection, ADDR, PORT):
    """ Establish a connection with the web client """

    print "Incoming Connection: ", ADDR, ' on ', PORT
    print "WAITING FOR INCOMING CONNECTION TO BE MADE"

    incoming_connection.sock = socket.socket()
    incoming_connection.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    incoming_connection.sock.bind((ADDR, PORT))
    incoming_connection.sock.listen(1)

    # Handle Request
    incoming_connection.conn, incoming_connection.addr = incoming_connection.sock.accept()
    print 'Incoming Connection made to: ', incoming_connection.addr
    data = incoming_connection.conn.recv(1024)
    print 'Received Msg: ', data
    response = create_handshake_resp(data)
    incoming_connection.conn.sendto(response, incoming_connection.addr)

    # Create a lock object
    incoming_connection.lock = threading.Lock()

    return incoming_connection

def close_outgoing_connection(outgoing_connection):
    """ Close Outgoing Connection """

    outgoing_connection.conn.shutdown(2)
    outgoing_connection.sock.close()

    return 0

def close_incoming_connection(incoming_connection):
    """ Close Incoming Connection """

    print 'Closing Client:', incoming_connection.addr

    incoming_connection.lock.acquire()
    incoming_connection.addr = ''
    incoming_connection.sock.close()
    incoming_connection.lock.release()

    return 0











import os, sys
import math

from bamboo.rdkit_utils.mol import check_pdb_readable
from bamboo.rdkit_utils.coords.utils import get_atomic_equivalences
from bamboo.wrappers.symutils import map_to_reference_using_symmetry

def calculate_rmsd_between_fragged_files_using_symmetry(file1, file2, delete_structures=True):
    """Take a pair of homologous fragmented molecule files and return a list of the rmsds between the paired fragments in each molecule, minimising over equivalent models"""

    # Map file2 to file1 using symmetry
    tempfile = map_to_reference_using_symmetry(refpdb=file1, movpdb=file2, pdbout=file2.replace('.pdb','.temp'))
    # Calculate the centroid as normal
    rmsds = calculate_rmsd_between_fragged_files(file1, tempfile)
    # Remove tempfiles and return
    if delete_structures: os.remove(tempfile)
    return rmsds

def calculate_rmsd_between_fragged_files(file1, file2):
    """Take a pair of homologous fragmented molecule files and return a list of the rmsds between the paired fragments in each molecule"""

    mol1 = check_pdb_readable(file1)
    mol2 = check_pdb_readable(file2)

    return calculate_rmsd_between_fragged_mols(mol1, mol2)

def calculate_rmsd_between_fragged_mols(mol1, mol2):
    """Take a pair of homologous fragmented molecules mols and return a list of the rmsds between the paired fragments in each molecule"""

    # TODO Extend this so it loops over conformers
    conf1 = mol1.GetConformer()
    conf2 = mol2.GetConformer()

    # Get the atom pairings between mol1, mol2 [[(mol1atm,mol2atm),...],...]
    try:
        match_patterns = get_atomic_equivalences(mol1, mol2)
    except EqualityError:
        raise

    # Assert that there is only one residue present (albeit with different insertion codes)
    assert len(list(set([at.GetMonomerInfo().GetChainId() for at in mol1.GetAtoms()])))==1, 'MORE THAN ONE CHAIN PRESENT IN MOL!'
    assert len(list(set([at.GetMonomerInfo().GetResidueName() for at in mol1.GetAtoms()])))==1, 'MORE THAN ONE RESIDUE NAME PRESENT IN MOL!'
    assert len(list(set([at.GetMonomerInfo().GetResidueNumber() for at in mol1.GetAtoms()])))==1, 'MORE THAN ONE RESIDUE NUMBER PRESENT IN MOL!'

    # Collect the fragments for mol1
    mol1_inscodes = sorted(list(set([at.GetMonomerInfo().GetInsertionCode() for at in mol1.GetAtoms()])))

    match_pattern_fragment_rmsds = []

    # Iterate through the different equivalent ways of matching the atoms between the molecules
    for match_pat in match_patterns:
        # Enable us to translate between mol1 and mol2
        match_dict = dict(match_pat)

        # RMSDs between fragments for this matching patterns
        fragment_rmsds = []
        total_dev_sq = []
        # List of the equivalent inscodes in mol2
        mol2_inscodes = []

        # Iterate through and form and compare the fragments for mol1 and mol2
        for mol1_inscode in mol1_inscodes:
            # Get all of the atoms in mol1 with the inscode == `inscode`
            mol1_frag_atms = [at for at in mol1.GetAtoms() if at.GetMonomerInfo().GetInsertionCode() == mol1_inscode]
            # Get the equivalent atoms in mol2
            mol2_frag_atms = [mol2.GetAtomWithIdx(match_dict[at.GetIdx()]) for at in mol1_frag_atms]

            # Get the equivalent mol2 inscode for this mol1 inscode
            mol2_inscodes.append(list(set([at.GetMonomerInfo().GetInsertionCode() for at in mol2_frag_atms]))[0])

            # Sum of length sqs of vectors joining atoms
            dists_sq = []
            for at1, at2 in zip(mol1_frag_atms, mol2_frag_atms):
                p1 = conf1.GetAtomPosition(at1.GetIdx())
                p2 = conf2.GetAtomPosition(at2.GetIdx())
                # Distance between points
                d = p1.Distance(p2)
                dists_sq.append(d**2)

            # Calculate rmsd for fragment and append
            rmsd = math.sqrt(sum(dists_sq)/len(dists_sq))
            fragment_rmsds.append(rmsd)

            # Keep a running total to enable a global rmsd to be calculated
            total_dev_sq.extend(dists_sq)

        # Calculate whole-molecule rmsd to allow for comparison between match_patterns
        total_rmsd = math.sqrt(sum(total_dev_sq)/len(total_dev_sq))

        # Record the fragment mappings, the fragment rmsds and the global rmsd
        match_pattern_fragment_rmsds.append((mol1_inscodes, mol2_inscodes, fragment_rmsds, total_rmsd))

    if not match_pattern_fragment_rmsds:
        raise Exception('No RMSDS calculated between fragments!')

    # Find the minimum of the total rmsds to find the optimum mapping
    min_rmsd = min([r[3] for r in match_pattern_fragment_rmsds])

    # Get those associated scores
    minimising_match_pattern = [r for r in match_pattern_fragment_rmsds if r[3] == min_rmsd][0]

    return zip(*minimising_match_pattern[0:3])




import os

from bamboo.wrappers.ligand_builders import BuilderObject

class WritedictObject(BuilderObject):

    def _custom_init(self):
        self.program = 'writedict'
        self.name = 'writedict'

    def _create_program_arguments(self, ligsmile, outdir, outfile, flags):
        """Prepare the arguments for Writedict to build the ligand"""

        # Create input smiles file
        self.infile = self.outtemplate+'.smi'
        # Write smile file
        with open(self.infile, 'w') as fh:
            fh.write(ligsmile)

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = ['-in',self.infile,'-out',self.outtemplate]
        cmd_line_args.extend(cmd_line_flags)
        # Form Std Input
        std_inpt_args = []
        std_inpt_args.extend(std_inpt_flags)

        return cmd_line_args, std_inpt_args

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = []
        std_inpt_flags = []

        if 'addHydrogens' in flags:
            raise SystemExit('AAAHHH - FLAG NOT PROGRAMMED FOR WRITEDICT - `noHydrogens`')
        if 'noHydrogens' in flags:
            raise SystemExit('AAAHHH - FLAG NOT PROGRAMMED FOR WRITEDICT - `noHydrogens`')
        # Change Ligand Name
        if 'correctNames' in flags:
            pass
        else:
            cmd_line_flags.append('-nolookup')

        return cmd_line_flags, std_inpt_flags



import os,sys,re

from bamboo.common.file import FileObj
from bamboo.macro.molecule import MacroMol

################################################################################################################
### CLASSES                                                                                                    #
################################################################################################################

class PdbMeta(object):
    """Object to hold the meta data for a PDB file"""

    def __init__(self):
        self.reslow = None
        self.reshigh = None
        self.spacegroup = None
        self.spacegroupno = None
        self.cell = None
        self.cryst = None
        self.rfree = None
        self.rwork = None
    def __str__(self):
        return str(self.__dict__)

class BFactorSummary(object):
    """Class for summarising the b-factors of a residue"""
    def __init__(self, chain, resid, main_bav, main_brms, side_bav, side_brms, all_bav, all_brms):
        self.chain = chain
        self.resid = resid
        self.main_bav  = main_bav
        self.main_brms = main_brms
        self.side_bav  = side_bav
        self.side_brms = side_brms
        self.all_bav   = all_bav
        self.all_brms  = all_brms

class PdbSummary(object):
    """Class for summarising an PDBfile"""
    def __init__(self, pdb_file, parse=False):
        self.path = os.path.abspath(pdb_file)
        self.file = FileObj(pdb_file)
        # Read the pdb_file and create a summary
        self.summary = get_pdb_summary_dict(pdb_file)
        # Get some Meta
        self.data = PdbMeta()
        self.data.reslow        = self.summary['reslow']        if self.summary['reslow']       else None
        self.data.reshigh       = self.summary['reshigh']       if self.summary['reshigh']      else None
        self.data.spacegroup    = self.summary['spacegroup']    if self.summary['spacegroup']   else None
        self.data.spacegroupno  = self.summary['spacegroupno']  if self.summary['spacegroupno'] else None
        self.data.cell          = self.summary['cell']
        self.data.cryst         = dict(zip(['A','B','C','ALPHA','BETA','GAMMA','SG'],self.data.cell+[self.data.spacegroup]))
        self.data.a,     self.data.b,    self.data.c     = self.summary['cell'][0:3]
        self.data.alpha, self.data.beta, self.data.gamma = self.summary['cell'][3:6]
        # Create MacroMol object
        if parse:
            self.macromol = MacroMol(pdbpath)
        else:
            self.macromol = None

################################################################################################################
### FUNCTIONS                                                                                                  #
################################################################################################################

def get_pdb_summary_dict(pdb_file):
    """Get a pdb summary from the header information"""

    # Extract contents of PDB
    pdbcontents = open(pdb_file, 'r').read()

    # Create empty dict to contain the summary
    summary = {}

    # Get the Refinement High Resolution Limit
    regex = re.compile('\nREMARK.*RESOLUTION RANGE HIGH \(ANGSTROMS\) :(.*)\n')
    matches = regex.findall(pdbcontents)
#    assert matches, 'No High Resolution Limit found for {!s}'.format(pdb_file)
    assert len(matches)<2, 'Too many matching lines found for High Res Limit in PDBFile {!s}\n\t{!s}'.format(pdb_file,matches)
    summary['reshigh'] = float(matches[0]) if matches else None

    # Get the Refinement Low Resolution Limit
    regex = re.compile('\nREMARK.*RESOLUTION RANGE LOW  \(ANGSTROMS\) :(.*)\n')
    matches = regex.findall(pdbcontents)
#    assert matches, 'No Low Resolution Limit found for {!s}'.format(pdb_file)
    assert len(matches)<2, 'Too many matching lines found for Low Res Limit in PDBFile {!s}\n\t{!s}'.format(pdb_file,matches)
    summary['reslow'] = float(matches[0]) if matches else None

    # Get the R Free
    regex = re.compile('\nREMARK.*\ \ FREE R VALUE\ *:(.*)\n')
    matches = regex.findall(pdbcontents)
#    assert matches, 'No R FREE found for {!s}'.format(pdb_file)
    assert len(matches)<2, 'Too many matching lines found for R FREE in PDBFile {!s}\n\t{!s}'.format(pdb_file,matches)
    summary['rfree'] = float(matches[0]) if matches else None

    # Get the R Work
    regex = re.compile('\nREMARK.*\ \ R VALUE.*\(WORKING SET\) :(.*)\n')
    matches = regex.findall(pdbcontents)
#    assert matches, 'No R WORK found for {!s}'.format(pdb_file)
    assert len(matches)<2, 'Too many matching lines found for R WORK in PDBFile {!s}\n\t{!s}'.format(pdb_file,matches)
    summary['rwork'] = float(matches[0]) if matches else None

    # Get the Cryst Line
    regex = re.compile('\n(CRYST1.*)\n')
    matches = regex.findall(pdbcontents)
    assert matches, 'No CRYST LINE found for {!s}'.format(pdb_file)
    assert len(matches)<2, 'Too many matching lines found for CRYST LINE in PDBFile {!s}\n\t{!s}'.format(pdb_file,matches)
    assert len(matches[0]) > 54, 'Cryst Line is not long enough in PDBFile {!s}\n\t{!s}'.format(pdb_file, matches)
    summary['cell'] = map(float,[matches[0][6:15], matches[0][15:24], matches[0][24:33], matches[0][33:40], matches[0][40:47], matches[0][47:54]])
    summary['spacegroup'] = matches[0][55:66].strip()
    summary['spacegroupno'] = None

    return summary

def get_pdb_cryst_line(pdb_file):
    """Get the cryst Line from the PDB File"""

    cryst = [l for l in open(pdb_file,'r').readlines() if l.startswith('CRYST')]
    # There should only be one CRYST line
    if len(cryst)!=1:
        return ('More than one cryst line found',cryst)
    else:
        cryst = cryst[0]
    # Check for validity of CRYST line
    if not cryst[0:6]=='CRYST1':
        return ('CRYST line does not begin with \'CRYST1\'',cryst)
    elif len(cryst)<56:
        return ('CRYST line is too short for PDB format',cryst)
    # Unit Cell Dimensions
    a = float(cryst[6:15])
    b = float(cryst[15:24])
    c = float(cryst[24:33])
    # Unit Cell Angles
    alpha = float(cryst[33:40])
    beta  = float(cryst[40:47])
    gamma = float(cryst[47:54])
    # Space Group and Z Value
    spacegroup = cryst[55:66].strip()
    if len(cryst) > 66:
        z_value = cryst[66:70]
    else:
        z_value = ''

    # Format
    cols = ['A', 'B', 'C', 'ALPHA', 'BETA', 'GAMMA', 'SG', 'Z']
    cryst = [a,b,c,alpha,beta,gamma,spacegroup,z_value]
    # Turn into dict with Column Headings as keys
    returncryst = dict(zip(cols,cryst))

    return (None,returncryst)




import os, sys, copy

import libtbx.phil

import iotbx.pdb
import iotbx.ccp4_map

import cctbx.maptbx
import cctbx.sgtbx
import iotbx.map_tools

from scitbx.array_family import flex

from giant.structure.align import align_chains

from mmtbx.maps.superpose import mask_grid, generate_p1_box

#######################################

blank_arg_prepend = None

master_phil = libtbx.phil.parse("""
align {
    reference = None
        .type = path
    chain = None
        .type = str
        .help = 'The chain to be aligned to'
    align_all_chains = False
        .type = bool
        .help = 'Align all chains in pdb to reference'
}
input {
    align_pdb = None
        .type = path
        .help = 'The pdb to use for alignment'
    pdb = None
        .type = path
        .multiple = True
    map = None
        .type = path
        .multiple = True
}
output {
    out_dir = ./
        .type = path
    pdb_suffix = '.aligned.pdb'
        .type = str
    pdb_prefix = ''
        .type = str
    map_suffix = '.aligned.map'
        .type = str
    map_prefix = ''
        .type = str
}
""")

#######################################

def run(params):

    assert params.align.reference, 'A REFERENCE PDB MUST BE PROVIDED'
    assert params.input.pdb or params.input.align_pdb, 'A PDB FILE TO BE ALIGNED MUST BE PROVIDED'
    if not params.input.align_pdb:
        params.input.align_pdb = params.input.pdb[0]
        print 'SELECTING FIRST PDB FOR ALIGNING: {}'.format(params.input.align_pdb)

    # Read in the reference structure
    ref_inp = iotbx.pdb.hierarchy.input(params.align.reference)
    if params.align.chain:
        ref_chain = [c for c in ref_inp.hierarchy.chains() if c.id == params.align.chain][0]
    else:
        ref_chain = [c for c in ref_inp.hierarchy.chains()][0]

    print 'ALIGNING TO CHAIN {!s} OF THE REFERENCE STRUCTURE'.format(ref_chain.id)

    # Read in the structure to be aligned
    align_inp = iotbx.pdb.hierarchy.input(params.input.align_pdb)

    # Build up a list of rotation matrices and write out transformed maps
    for align_chain in align_inp.hierarchy.chains():
        if not align_chain.is_protein():
            continue
        if (not params.align.align_all_chains) and (not align_chain.id==ref_chain.id):
            continue

        print 'GENERATING ALIGNMENT'

        # Align to the reference chain
        lsq = align_chains(mov_chain=align_chain, ref_chain=ref_chain)
        # Extract the rotation matrix
        rt = lsq.rt()

        # Loop through the pdbs and transform each one
        for pdb_file in params.input.pdb:

            print 'TRANSFORMING:', pdb_file
            # Read in the structure to be aligned
            mov_inp = iotbx.pdb.hierarchy.input(pdb_file)
            # Generate output filename
            output_pdb = os.path.join(params.output.out_dir, params.output.pdb_prefix + os.path.basename(pdb_file) + '.chain{!s}'.format(align_chain.id) + params.output.pdb_suffix)
            # Transform coordinates and write
            mov_copy = mov_inp.hierarchy.deep_copy()
            mov_copy.atoms().set_xyz(rt*mov_copy.atoms().extract_xyz())
            mov_copy.write_pdb_file(output_pdb)

        # Loop through the maps and transform each one
        for map_file in params.input.map:

            print 'TRANSFORMING:', map_file
            # Read in the map to be aligned
            mov_map = iotbx.ccp4_map.map_reader(map_file)
            # Unpad the input map
            gridding_first = mov_map.data.origin()
            gridding_last = tuple(flex.int(mov_map.data.all())-2)
            mov_data = cctbx.maptbx.copy(mov_map.data.as_double(), first=gridding_first, last=gridding_last)
            # Generate output filename
            output_map = os.path.join(params.output.out_dir, params.output.map_prefix + os.path.basename(map_file) + '.chain{!s}'.format(align_chain.id) + params.output.map_suffix)
            assert not os.path.exists(output_map), 'OUTPUT MAP ALREADY EXISTS: {!s}'.format(output_map)
            # Rotate the map
            rot_map_data = cctbx.maptbx.rotate_translate_map(
                unit_cell=mov_map.unit_cell(),
    #            map_data=mov_map.data.as_double(),
                map_data=mov_data,
                rotation_matrix=rt.inverse().r.elems,
                translation_vector=rt.inverse().t.elems)
            # Write the map
            iotbx.ccp4_map.write_ccp4_map(
                file_name=output_map,
                unit_cell=mov_map.unit_cell(),
                space_group=cctbx.sgtbx.space_group_info(number=mov_map.space_group_number).group(),
                map_data=rot_map_data,
                labels = flex.std_string(['rotated map. aligned chain {!s} to chain {!s}.'.format(align_chain.id, ref_chain.id)])  )

#        # New change map
#        fake_symm = generate_p1_box(ref_inp.hierarchy, buffer=10.0)
#        xray_structure = ref_inp.hierarchy.extract_xray_structure(crystal_symmetry=fake_symm)
#        f_calc = xray_structure.structure_factors(d_min=2).f_calc()
#        fake_map = f_calc.fft_map(resolution_factor=0.33)
#
#        map_data_superposed = cctbx.maptbx.superpose_maps(
#            unit_cell_1        = mov_map.unit_cell(),
#            unit_cell_2        = fake_symm.unit_cell(),
#            map_data_1         = mov_map.data.as_double(),
#            n_real_2           = fake_map.n_real(),
#            rotation_matrix    = rt.inverse().r.elems,
#            translation_vector = rt.inverse().t.elems)
#
#        map_data_superposed = mask_grid(
#            xrs      = xray_structure,
#            buffer   = 10,
#            map_data = map_data_superposed,
#            n_real   = fake_map.n_real())
#
#        iotbx.map_tools.write_ccp4_map(
#            sites_cart=xray_structure.sites_cart(),
#            unit_cell=fake_symm.unit_cell(),
#            map_data=map_data_superposed,
#            n_real=fake_map.n_real(),
#            file_name=output_map,
#            buffer=10)

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os

from bamboo.common.command import CommandManager
from bamboo.maps.convert import convert_mtz_to_map
from bamboo.wrappers.maputils import mask_map

def create_masked_map_from_file_objects(mtzobj,pdbobj,mapout=None,maptype='2FOFC',force=False):
    """Takes the ligand described by ligobj and masks against the density from pdbobj & mtzobj"""

    # oriented and positioned structure of ligand only
    pdb_file = pdbobj.path
    # reflection data path
    mtz_file = mtzobj.path
    # Create the map
    fft, masker = create_masked_map(mtzin=mtz_file, maskpdb=pdb_file, mapout=mapout, maptype=maptype)

    return fft, masker

def create_masked_map(mtzin, maskpdb, mapout=None, maptype='2FOFC', border=2):
    """Creates density from mtzin, and masks around maskpdb - NOTE THAT MASKPDB MUST BE ALIGNED WITH MTZIN"""

    # Create mapfile if not given
    if not mapout:
        mapout = os.path.splitext(maskpdb)[0]+'.masked.'+maptype+'.ccp4'
    # Temporary (unmasked) map
    tempmap = mapout+'.nomask'
    # Convert the inital MAT to a MAP file
    fft = convert_mtz_to_map(mtzfile=mtzin, mapfile=tempmap, maptype=maptype)
    # Mask the map
    masker = mask_map(mapin=tempmap,maskpdb=maskpdb,mapout=mapout,border=border)
    # Remove the temporary mapfile
    os.remove(tempmap)
    # Return Command Managers for flexible handling of out & err
    return fft, masker


